# COPYRIGHT 2022 MONOOCLE INSIGHTS
# ALL RIGHTS RESERVED
# UNAUTHORIZED USE, REPRODUCTION, OR
# DISTRIBUTION STRICTLY PROHIBITED
#
# AUTHOR: Sean O'Malley and Raymond Deiotte
# CREATION DATE: 20230802
# LAST UPDATED: 20230821

### Import Packages/Dependencies ###

# connection
import sys
# basic
from datetime import datetime
import paramiko
from io import BytesIO
import pandas as pd
import requests

# import status package
#FIXME Use for PyPi packaging
# from care_query.care_query_status import CQStatus


### Query Data Class ###
class CareQueryData:

    def __init__(self, query, features, show = None, sftp_key = None):

        #FIXME
        # self.endpoint = 'http://localhost:5533/get_care_query'
        # self.endpoint = 'http://ec2-3-133-102-108.us-east-2.compute.amazonaws.com:5533/get_care_query'
        self.endpoint = 'http://CQBackendLB-1880313232.us-east-2.elb.amazonaws.com:5533/get_care_query'
        self.apikey = ''
        self.length = None
        self.features = features
        self.query = query

        # pass sftp key from primary class if already specified
        if sftp_key != None:
            self.sftp_key = sftp_key 

        # extract needed features
        self.token = self.query["token"]

        self.sftpmap = {
            'tendo.com':'tendo-user',
            'monocleinsights.com': 'monocle-user',
            'lifeonbelay.com' : 'lob-user',
            'summusglobal.com':'summus-user',
            'automationandinsight.com': 'manhattan-user'
        }
        
        if show != False:
            self.summary = self.summarize(show)

    def summarize(self, show = None):
        
        # un-nest
        features = self.query

        # show criterion
        if str(features["query"]) in ["encounter", "visit", "journey", "episode"]:
        
            query_text = str(features["query"]).replace("_"," ")
            min_date_text = str(features["min_date"])
            max_date_text = str(features["max_date"])
            gender_text = str(features["gender"]).replace("[","").replace("]","").replace(","," and ").replace("m","Males").replace("f","Females")
            age_min_text = features["age_min"]
            age_max_text = features["age_max"]

            if features["limit"] != None:
                limit_text = "Query will return the top " + "{:,.0f}".format(features["limit"]) + " rows meeting the query criterion."
            else:
                limit_text = "No Volume Limit: query will return ALL results"

            # dynamic warning list
            if features["warnings"] != None:
                
                try:
                    warnings_list_text2a = features["warnings"][0]
                    warnings_list_text = f"""
                    Query Logic:

                    {warnings_list_text2a}

                    ---------------"""
                except:
                    next
                try:
                    warnings_list_text2b = features["warnings"][1]
                    warnings_list_text = f"""
                    Query Logic:

                    {warnings_list_text2a}
                    {warnings_list_text2b}

                    ---------------"""
                except:
                    next
                try:
                    warnings_list_text2c = features["warnings"][2]
                    warnings_list_text = f"""
                    Query Logic:

                    {warnings_list_text2a}
                    {warnings_list_text2b}
                    {warnings_list_text2c}

                    ---------------"""
                except:
                    next
                try:
                    warnings_list_text2d = features["warnings"][3]
                    warnings_list_text = f"""
                    Query Logic:

                    {warnings_list_text2a}
                    {warnings_list_text2b}
                    {warnings_list_text2c}
                    {warnings_list_text2d}

                    ---------------"""
                except:
                    next
                try:
                    warnings_list_text2e = features["warnings"][4]
                    warnings_list_text = f"""
                    Query Logic:

                    {warnings_list_text2a}
                    {warnings_list_text2b}
                    {warnings_list_text2c}
                    {warnings_list_text2d}
                    {warnings_list_text2e}

                    ---------------"""
                except:
                    next
                try:
                    warnings_list_text2f = features["warnings"][5]
                    warnings_list_text = f"""
                    Query Logic:

                    {warnings_list_text2a}
                    {warnings_list_text2b}
                    {warnings_list_text2c}
                    {warnings_list_text2d}
                    {warnings_list_text2e}
                    {warnings_list_text2f}
                    ...
                    ---------------"""
                except:
                    next
            else:
                warnings_list_text = "No Query Warnings Flagged"

            # state
            if (features["state"] != None):
                state_text = f"""from any of the below states:
                    {str(features["state"]).replace(", "," or ").replace("[","").replace("]","")}"""
            else:
                state_text = ""

            # short zip
            if (features["short_zip"] != None):
                if (features["state"] != None):
                    zip_text = f"""or any of the below zip codes:
                    {str(features["short_zip"]).replace(", "," or ").replace("[","").replace("]","")}"""
                else:
                    zip_text = f"""from any of the below zip codes:
                    {str(features["short_zip"]).replace(", "," or ").replace("[","").replace("]","")}"""
            else:
                zip_text = ""

            # if non, then all
            if (zip_text == "") & (state_text == ""):
                state_text = "Any/All Geographies Available to User's Access Plan Will be Queried"
                
            # create operand based on greedy preferences 
            if features["greedy"] != False:
                operand = "--- OR ---"
            else:
                operand = "--- AND ---"

            # visit type
            if features["visit_type"] != None:
                if any(isinstance(el, list) for el in features["visit_type"]):
                    try:
                        visit_type_text = f"""Visit Type(s):
                    {str(features["visit_type"][0]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        visit_type_text = f"""Visit Type(s):
                    {str(features["visit_type"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["visit_type"][1]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        visit_type_text = f"""Visit Type(s):
                    {str(features["visit_type"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["visit_type"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["visit_type"][2]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        visit_type_text = f"""Visit Type(s):
                    {str(features["visit_type"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["visit_type"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["visit_type"][2]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["visit_type"][3]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:    
                        visit_type_text = f"""Visit Type(s):
                    {str(features["visit_type"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["visit_type"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["visit_type"][2]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["visit_type"][3]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["visit_type"][4]).replace(", "," or ").replace("[","").replace("]","")}
                    ..."""
                    except:
                        next
                else:
                    visit_type_text = f"""Visit Type(s):
                    {str(features["visit_type"]).replace(", "," or ").replace("[","").replace("]","")}"""
            else:
                visit_type_text = "Any/All Corresponding Visit Type(s)"

            # payor channels
            if features["payor_channel"] != None:
                if any(isinstance(el, list) for el in features["payor_channel"]):
                    try:
                        payor_channel_text = f"""Payor Channel(s):
                    {str(features["payor_channel"][0]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        payor_channel_text = f"""Payor Channel(s):
                    {str(features["payor_channel"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["payor_channel"][1]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        payor_channel_text = f"""Payor Channel(s):
                    {str(features["payor_channel"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["payor_channel"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["payor_channel"][2]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        payor_channel_text = f"""Payor Channel(s):
                    {str(features["payor_channel"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["payor_channel"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["payor_channel"][2]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["payor_channel"][3]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:    
                        payor_channel_text = f"""Payor Channel(s):
                    {str(features["payor_channel"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["payor_channel"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["payor_channel"][2]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["payor_channel"][3]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["payor_channel"][4]).replace(", "," or ").replace("[","").replace("]","")}
                    ..."""
                    except:
                        next
                else:
                    payor_channel_text = f"""Payor Channel(s):
                    {str(features["payor_channel"]).replace(", "," or ").replace("[","").replace("]","")}"""
            else:
                payor_channel_text = "Any/All Corresponding Payor Channel(s)"

            # revenue center codes
            if features["revenue_center_code"] != None:
                if any(isinstance(el, list) for el in features["revenue_center_code"]):
                    try:
                        revenue_center_code_text = f"""Revenue Center Code(s):
                    {str(features["revenue_center_code"][0]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        revenue_center_code_text = f"""Revenue Center Code(s):
                    {str(features["revenue_center_code"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["revenue_center_code"][1]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        revenue_center_code_text = f"""Revenue Center Code(s):
                    {str(features["revenue_center_code"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["revenue_center_code"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["revenue_center_code"][2]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        revenue_center_code_text = f"""Revenue Center Code(s):
                    {str(features["revenue_center_code"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["revenue_center_code"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["revenue_center_code"][2]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["revenue_center_code"][3]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:    
                        revenue_center_code_text = f"""Revenue Center Code(s):
                    {str(features["revenue_center_code"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["revenue_center_code"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["revenue_center_code"][2]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["revenue_center_code"][3]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["revenue_center_code"][4]).replace(", "," or ").replace("[","").replace("]","")}
                    ..."""
                    except:
                        next
                else:
                    revenue_center_code_text = f"""Revenue Center Code(s):
                    {str(features["revenue_center_code"]).replace(", "," or ").replace("[","").replace("]","")}"""
            else:
                revenue_center_code_text = "Any/All Corresponding Revenue Center Code(s)"

            # place of service
            if features["place_of_service"] != None:
                if any(isinstance(el, list) for el in features["place_of_service"]):
                    try:
                        place_of_service_text = f"""Place(s) of Service:
                    {str(features["place_of_service"][0]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        place_of_service_text = f"""Place(s) of Service:
                    {str(features["place_of_service"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["place_of_service"][1]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        place_of_service_text = f"""Place(s) of Service:
                    {str(features["place_of_service"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["place_of_service"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["place_of_service"][2]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        place_of_service_text = f"""Place(s) of Service:
                    {str(features["place_of_service"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["place_of_service"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["place_of_service"][2]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["place_of_service"][3]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:    
                        place_of_service_text = f"""Place(s) of Service:
                    {str(features["place_of_service"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["place_of_service"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["place_of_service"][2]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["place_of_service"][3]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["place_of_service"][4]).replace(", "," or ").replace("[","").replace("]","")}
                    ..."""
                    except:
                        next
                else:
                    place_of_service_text = f"""Place(s) of Service:
                    {str(features["place_of_service"]).replace(", "," or ").replace("[","").replace("]","")}"""
            else:
                place_of_service_text = "Any/All Corresponding Place(s) of Service"

            # drg
            # if features["drg_code"] != None:
            #     if any(isinstance(el, list) for el in features["drg_code"]):
            #         try:
            #             drg_code_text = f"""DRGs - Diagnosis Related Group(s):
            #         {str(features["drg_code"][0]).replace(", "," or ").replace("[","").replace("]","")}"""
            #         except:
            #             next
            #         try:
            #             drg_code_text = f"""DRGs - Diagnosis Related Group(s):
            #         {str(features["drg_code"][0]).replace(", "," or ").replace("[","").replace("]","")}
            #         and
            #         {str(features["drg_code"][1]).replace(", "," or ").replace("[","").replace("]","")}"""
            #         except:
            #             next
            #         try:
            #             drg_code_text = f"""DRGs - Diagnosis Related Group(s):
            #         {str(features["drg_code"][0]).replace(", "," or ").replace("[","").replace("]","")}
            #         and
            #         {str(features["drg_code"][1]).replace(", "," or ").replace("[","").replace("]","")}
            #         and
            #         {str(features["drg_code"][2]).replace(", "," or ").replace("[","").replace("]","")}"""
            #         except:
            #             next
            #         try:
            #             drg_code_text = f"""DRGs - Diagnosis Related Group(s):
            #         {str(features["drg_code"][0]).replace(", "," or ").replace("[","").replace("]","")}
            #         and
            #         {str(features["drg_code"][1]).replace(", "," or ").replace("[","").replace("]","")}
            #         and
            #         {str(features["drg_code"][2]).replace(", "," or ").replace("[","").replace("]","")}
            #         and
            #         {str(features["drg_code"][3]).replace(", "," or ").replace("[","").replace("]","")}"""
            #         except:
            #             next
            #         try:    
            #             drg_code_text = f"""DRGs - Diagnosis Related Group(s):
            #         {str(features["drg_code"][0]).replace(", "," or ").replace("[","").replace("]","")}
            #         and
            #         {str(features["drg_code"][1]).replace(", "," or ").replace("[","").replace("]","")}
            #         and
            #         {str(features["drg_code"][2]).replace(", "," or ").replace("[","").replace("]","")}
            #         and
            #         {str(features["drg_code"][3]).replace(", "," or ").replace("[","").replace("]","")}
            #         and
            #         {str(features["drg_code"][4]).replace(", "," or ").replace("[","").replace("]","")}
            #         ..."""
            #         except:
            #             next
            #     else:
            #         drg_code_text = f"""DRGs - Diagnosis Related Group(s):
            #         {str(features["drg_code"]).replace(", "," or ").replace("[","").replace("]","")}"""
            # else:
            #     drg_code_text = "Any/All Corresponding DRGs - Diagnosis Related Group(s)"

            # diagnosis code
            if features["diag_code"] != None:
                if any(isinstance(el, list) for el in features["diag_code"]):
                    try:
                        diag_code_text = f"""Diagnosis Code(s):
                    {str(features["diag_code"][0]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        diag_code_text = f"""Diagnosis Code(s):
                    {str(features["diag_code"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["diag_code"][1]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        diag_code_text = f"""Diagnosis Code(s):
                    {str(features["diag_code"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["diag_code"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["diag_code"][2]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        diag_code_text = f"""Diagnosis Code(s):
                    {str(features["diag_code"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["diag_code"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["diag_code"][2]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["diag_code"][3]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:    
                        diag_code_text = f"""Diagnosis Code(s):
                    {str(features["diag_code"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["diag_code"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["diag_code"][2]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["diag_code"][3]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["diag_code"][4]).replace(", "," or ").replace("[","").replace("]","")}
                    ..."""
                    except:
                        next
                else:
                    diag_code_text = f"""Diagnosis Code(s):
                    {str(features["diag_code"]).replace(", "," or ").replace("[","").replace("]","")}"""
            else:
                diag_code_text = "Any/All Corresponding Diagnosis Codes"

            # procedure code
            if features["proc_code"] != None:
                if any(isinstance(sub, list) for sub in features["proc_code"]):
                    try:
                        proc_code_text = f"""Procedure(s):
                    {str(features["proc_code"][0]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        proc_code_text = f"""Procedure(s):
                    {str(features["proc_code"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["proc_code"][1]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        proc_code_text = f"""Procedure(s):
                    {str(features["proc_code"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["proc_code"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["proc_code"][2]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        proc_code_text = f"""Procedure(s):
                    {str(features["proc_code"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["proc_code"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["proc_code"][2]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["proc_code"][3]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:    
                        proc_code_text = f"""Procedure(s):
                    {str(features["proc_code"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["proc_code"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["proc_code"][2]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["proc_code"][3]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["proc_code"][4]).replace(", "," or ").replace("[","").replace("]","")}
                    ..."""
                    except:
                        next
                else:
                    proc_code_text = f"""Procedure(s):
                    {str(features["proc_code"]).replace(", "," or ").replace("[","").replace("]","")}"""
            else:
                    proc_code_text = "Any/All Corresponding Procedure Codes"

            # taxonomy code
            if features["taxonomy_code"] != None:
                if any(isinstance(sub, list) for sub in features["taxonomy_code"]):
                    try:
                        taxonomy_code_text = f"""Taxonomy Code(s):
                    {str(features["taxonomy_code"][0]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        taxonomy_code_text = f"""Taxonomy Code(s):
                    {str(features["taxonomy_code"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["taxonomy_code"][1]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        taxonomy_code_text = f"""Taxonomy Code(s):
                    {str(features["taxonomy_code"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["taxonomy_code"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["taxonomy_code"][2]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        taxonomy_code_text = f"""Taxonomy Code(s):
                    {str(features["taxonomy_code"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["taxonomy_code"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["taxonomy_code"][2]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["taxonomy_code"][3]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:    
                        taxonomy_code_text = f"""Taxonomy Code(s):
                        {str(features["taxonomy_code"][0]).replace(", "," or ").replace("[","").replace("]","")}
                        and
                        {str(features["taxonomy_code"][1]).replace(", "," or ").replace("[","").replace("]","")}
                        and
                        {str(features["taxonomy_code"][2]).replace(", "," or ").replace("[","").replace("]","")}
                        and
                        {str(features["taxonomy_code"][3]).replace(", "," or ").replace("[","").replace("]","")}
                        and
                        {str(features["taxonomy_code"][4]).replace(", "," or ").replace("[","").replace("]","")}
                        ..."""
                    except:
                        next
                else:
                    taxonomy_code_text = f"""Taxonomy Code(s):
                    {str(features["taxonomy_code"]).replace(", "," or ").replace("[","").replace("]","")}"""
            else:
                    taxonomy_code_text = "Any/All Corresponding (Specialty) Taxonomy Codes"

            # referring npis
            if features["ref_npi"] != None:
                if any(isinstance(el, list) for el in features["ref_npi"]):
                    try:
                        ref_npi_text = f"""Referring NPI(s):
                    {str(features["ref_npi"][0]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        ref_npi_text = f"""Referring NPI(s):
                    {str(features["ref_npi"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["ref_npi"][1]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        ref_npi_text = f"""Referring NPI(s):
                    {str(features["ref_npi"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["ref_npi"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["ref_npi"][2]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        ref_npi_text = f"""Referring NPI(s):
                    {str(features["ref_npi"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["ref_npi"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["ref_npi"][2]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["ref_npi"][3]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:    
                        ref_npi_text = f"""Referring NPI(s):
                    {str(features["ref_npi"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["ref_npi"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["ref_npi"][2]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["ref_npi"][3]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["ref_npi"][4]).replace(", "," or ").replace("[","").replace("]","")}
                            ..."""
                    except:
                        next
                else:
                    ref_npi_text = f"""Referring NPI(s):
                    {str(features["ref_npi"]).replace(", "," or ").replace("[","").replace("]","")}"""
            else:
                ref_npi_text = "Any/All Corresponding Referring NPI(s)"

            #  npis
            if features["npi"] != None:
                if any(isinstance(el, list) for el in features["npi"]):
                    try:
                        npi_text = f"""NPI(s) - National Provider Identifier(s):
                    {str(features["npi"][0]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        npi_text = f""":
                    {str(features["npi"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["npi"][1]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        npi_text = f"""NPI(s) - National Provider Identifier(s):
                    {str(features["npi"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["npi"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["npi"][2]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:
                        npi_text = f"""NPI(s) - National Provider Identifier(s):
                    {str(features["npi"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["npi"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["npi"][2]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["npi"][3]).replace(", "," or ").replace("[","").replace("]","")}"""
                    except:
                        next
                    try:    
                        npi_text = f"""NPI(s) - National Provider Identifier(s):
                    {str(features["npi"][0]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["npi"][1]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["npi"][2]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["npi"][3]).replace(", "," or ").replace("[","").replace("]","")}
                    and
                    {str(features["npi"][4]).replace(", "," or ").replace("[","").replace("]","")}
                            ..."""
                    except:
                        next
                else:
                    npi_text = f"""NPI(s) - National Provider Identifier(s):
                    {str(features["npi"]).replace(", "," or ").replace("[","").replace("]","")}"""
            else:
                npi_text = "Any/All Corresponding NPI(s) - National Provider Identifier(s)"

            self.summary = f"""
                    ---------------------------------------------------------------------------------------------------------
                    ---------------------------------------------------------------------------------------------------------
                    {query_text.title()} Query Summary
                    (specify 'show = False' within query method to suppress message)
                    ---------------------------------------------------------------------------------------------------------
                    ---------------------------------------------------------------------------------------------------------
                    {warnings_list_text} 

                    Query Volume Limit: 
                    {limit_text} 

                    ---------------

                    Query will return line-item granularity for {query_text}s from {min_date_text} to {max_date_text} that meet the following criterion...

                    Population: 

                    {gender_text} patients between the ages of {age_min_text} and {age_max_text} 

                    {state_text} 
                    {zip_text}

                    Specifications:

                    {visit_type_text}

                    {operand}

                    {payor_channel_text}

                    {operand}

                    {revenue_center_code_text}

                    {operand}

                    {place_of_service_text}

                    {operand}

                    {diag_code_text}

                    {operand}

                    {proc_code_text}

                    {operand}

                    {taxonomy_code_text}

                    {operand}

                    {ref_npi_text}

                    {operand}

                    {npi_text}

                    ---------------------------------------------------------------------------------------------------------
                    ---------------------------------------------------------------------------------------------------------
                            """

            if show == True:
                
                print(self.summary)

        else:

            print(features)

    def estimate(self, as_pandas = None):

        # update mode
        self.query.update({"mode":"estimate"})
        self.query.update({"timestamp":datetime.now().strftime('%Y%m%d_%H%M%S')})

        # execute to SFTP, inform user
        if self.query["query"] in ["journey", "visit", "episode", "encounter", "script-encounter", "script-journey"]:
            self.info = self.sendRequest("estimate")
            self.sftp_path = self.info["INFO"]["0"].split(": ")[1].replace(".","")
            self.query_id = self.info["STATUS"]["0"].split(": ")[1]
            if self.info != None:
                print("---")
                print("   ")
                print(f"""{self.query["query"]} query {self.query["mode"]} results will be returned to the SFTP location specified below. Users will be notified of query completion via email.""")
                print("   ")
                print(f"sftp_path: {self.sftp_path}")
                print(f"query_id: {self.query_id}")
                print("   ")
                print("*** Query status may be accessed via .queryStatus() method")
                print("*** Read data from SFTP location via .sftpToPandas() method - data <2M rows recommended")
                print("   ")
                print("---")
            return self.info
        # return locally
        else:
            self.estimate = self.sendRequest("estimate")

            # return as pandas dataframe
            if as_pandas != False:
                self.estimate = pd.DataFrame(self.estimate)
            return self.estimate
        
    def sample(self, as_pandas = None):

        # update mode
        self.query.update({"mode":"sample"})
        self.query.update({"timestamp":datetime.now().strftime('%Y%m%d_%H%M%S')})

        # execute to SFTP, inform user
        if self.query["query"] in ["journey", "visit", "episode", "encounter", "script-encounter", "script-journey"]:
            self.info = self.sendRequest("sample")
            self.sftp_path = self.info["INFO"]["0"].split(": ")[1].replace(".","")
            self.query_id = self.info["STATUS"]["0"].split(": ")[1]
            if self.info != None:
                print("---")
                print("   ")
                print(f"""{self.query["query"]} query {self.query["mode"]} results will be returned to the SFTP location specified below. You will be notified of query completion via email.""")
                print("   ")
                print(f"sftp_path: {self.sftp_path}")
                print(f"query_id: {self.query_id}")
                print("   ")
                print("*** Query status may be accessed via .queryStatus() method")
                print("*** Read data from SFTP location via .sftpToPandas() method - data <2M rows recommended")
                print("   ")
                print("---")
            return self.info
        # return locally
        else:
            self.sample = self.sendRequest("sample")

            # return as pandas dataframe
            if as_pandas != False:
                self.sample = pd.DataFrame(self.sample)
            return self.sample
        
    def execute(self, as_pandas = None):

        # update mode
        self.query.update({"mode":"execute"})
        self.query.update({"timestamp":datetime.now().strftime('%Y%m%d_%H%M%S')})

        # execute to SFTP, inform user
        if self.query["query"] in ["journey", "visit", "episode", "encounter", "script-encounter", "script-journey"]:
            self.info = self.sendRequest()
            self.sftp_path = self.info["INFO"]["0"].split(": ")[1].replace(".","")
            self.query_id = self.info["STATUS"]["0"].split(": ")[1]
            if self.info != None:
                print("---")
                print("   ")
                print(f"""{self.query["query"]} query results will be returned to the SFTP location specified below. You will be notified of query completion via email.""")
                print("   ")
                print(f"sftp_path: {self.sftp_path}")
                print(f"query_id: {self.query_id}")
                print("   ")
                print("*** Query status may be accessed via .queryStatus() method")
                print("*** Data may be accessed from SFTP via .sftpToPandas() method - data <2M rows recommended")
                print("   ")
                print("---")
            return self.info

        # execute to local, inform user
        else:
            self.data = self.sendRequest()

            # return as pandas dataframe
            if as_pandas != False:
                self.data = pd.DataFrame(self.data)

            return self.data

    def sendRequest(self, type = None):
        '''
        Method to send the request to the CQ API Endpoint
        @param data:
        @return:
        '''
        response = requests.post(self.endpoint, json=self.query)

        self.data = None
        #TODO - get the payload back r.data?
        if response.status_code == 200:
            if type == "estimate":
                if self.query["query"] not in ["journey", "visit", "episode", "encounter", "script-encounter", "script-journey"]:
                    print("---")
                    print(f'Successfully calculated the size of the query...')
                self.data = response.json()
            elif type == "sample":
                if self.query["query"] not in ["journey", "visit", "episode", "encounter", "script-encounter", "script-journey"]:
                    print("---")
                    print(f'Successfully made a sample of the query...')
                self.data = response.json()
            else:
                if self.query["query"] not in ["journey", "visit", "episode", "encounter", "script-encounter", "script-journey"]:
                    print("---")
                    print(f'Successfully executed the query...')
                self.data = response.json()
        else:
            if type == "estimate":
                print("---")
                print(f"Failed to execute the query: {response}")
            elif type == "sample":
                print("---")
                print(f"Failed to execute a sample of the query: {response}")
            else:
                print("---")
                print(f"Failed to execute the query: {response}")

        return self.data

    def queryStatus(self):

        tokenend = 'http://CQBackendLB-1880313232.us-east-2.elb.amazonaws.com:5533/get_care_query_status'
        # params = {}
        params = {'user': self.token,
                  'query': self.query_id}

        response = requests.post(tokenend, json=params)
        self.status = response.content.decode('utf-8')

        return f"Query Status: {self.status}"

    def sftpToPandas(self, sftp_key = None, sftp_path = None):
        
        """
        This method downloads parquet files from SFTP using public key authentication.
        
        Parameters:
            sftp_key (str):
                The sftp_key is the local location of the PEM file provided by the Monocle team. This file contains an encryption key that gives your machine access to the SFTP Location. This file needs to be saved locally, and that file location needs to be specified as the sftp_key as a string. 
                default: This can be set upon CareQuery instantiation or directly in this method
            sftp_path (str):
                The sftp_path is the remote file location where your query has been written to on the SFTP server.
                default: This is automatically writen to self.sftp_path when a query is made, but can be custom specified by the user

        """

        # references
        org = self.query['email'].split('@')[1].lower()
        customer = self.sftpmap[org]

        # prioritize
        self.hostname = 's-ff3fdf0d76104019a.server.transfer.us-east-2.amazonaws.com'
        if customer != None:
            self.customer = customer
        if sftp_key != None:
            self.sftp_key = sftp_key
        if sftp_path != None:
            self.sftp_path = sftp_path

        # Create an SSH client instance.
        ssh_client = paramiko.SSHClient()
        ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        # Connect to the SFTP server
        private_key = paramiko.RSAKey.from_private_key_file(self.sftp_key)
        ssh_client.connect(self.hostname, username=self.customer, pkey=private_key)

        # Create an SFTP client from the SSH client
        sftp_client = ssh_client.open_sftp()

        # List all files in the remote directory
        file_list = sftp_client.listdir(self.sftp_path)

        # Filter for Parquet files
        parquet_files = [file for file in file_list if file.endswith('.parquet')]

        dataframes = []
        for file in parquet_files:
            file_path = f"{self.sftp_path}/{file}"
            with sftp_client.open(file_path, 'rb') as remote_file:
                file_content = remote_file.read()
                df = pd.read_parquet(BytesIO(file_content))
                dataframes.append(df)

        # Concatenate all dataframes
        combined_df = pd.concat(dataframes, ignore_index=True)

        # Close SFTP and SSH connections
        sftp_client.close()
        ssh_client.close()

        return combined_df
    



# COPYRIGHT 2022 MONOOCLE INSIGHTS
# ALL RIGHTS RESERVED
# UNAUTHORIZED USE, REPRODUCTION, OR
# DISTRIBUTION STRICTLY PROHIBITED
#
# AUTHOR: Sean O'Malley and Raymond Deiotte
# CREATION DATE: 20230821
# LAST UPDATED: 20230821

### Import Dependencies ###

import re
# basic
import pandas as pd
from datetime import date, datetime

### Messy Data Helper ###

def ndcToNdc11(ndc):
    """
    Convert any formatted National Drug Code into NDC11
    """

    # Remove hyphens and spaces
    ndc_without_hyphens = ndc.replace("-", "").replace(" ", "")

    # Add leading zeros if necessary
    if len(ndc_without_hyphens) < 11:
        ndc_without_hyphens = ndc_without_hyphens.zfill(11)

    # Extract the first 11 characters
    ndc_11_code = ndc_without_hyphens[:11]

    return ndc_11_code

def finalJourneyClean(df):
    """
    On occasions of squirrely output, ensure data type and de-duplication for key features
    """
    try:
        df = df.drop_duplicates()
    except:
        next
    try:
        df["date"] = pd.to_datetime(df["date"])
    except:
        next
    try:
        df["smart_allowed"] = df["smart_allowed"].astype(float).round(2)
    except:
        next
    try:
        df["charge"] = df["charge"].astype(float).round(2)
    except:
        next
    try:
        df["proc_units"] = df["proc_units"].astype(float)
        df["rev_center_units"] = df["rev_center_units"].astype(float)
    except:
        next
        
    return df

### Data Type Helpers ###

def checkTypeMixed(x, x_name):
    if type(x) not in [int, str, list]:
        raise TypeError(f"TypeError: Datatype for '{x_name}' must be a string, int or list")
    
def checkTypeBool(x, x_name):
    if type(x) not in [bool]:
        raise TypeError(f"TypeError: Datatype for '{x_name}' must be a boolean")
    
def checkTypeInt(x, x_name):
    if type(x) not in [int]:
        raise TypeError(f"TypeError: Datatype for '{x_name}' must be a integer")
    
def checkTypeDate(x, x_name):
    if type(x) not in [date, datetime.date, datetime]:
        raise TypeError(f"TypeError: Datatype for '{x_name}' must be a date string or date object")

### Data Validation Helpers ###   

def checkNPI(param, name):
    """
    Error if NPI dimensions are invalid
    """
    if type(param) == str: # int to str
        if len(param) != 10:
            raise ValueError(f"ValueError: Invalid '{name}' length detected, must equal 10 digits and '{param}' is {len(param)}")
    elif type(param) == list: 
        if any(isinstance(el, list) for el in param): # if list of list
            naughty = list(set([item for sublist in param for item in sublist if len(item) != 10]))
            if len(naughty) > 0:
                raise ValueError(f"ValueError: Invalid '{name}' length detected, must equal 10 digits and {naughty} do not meet this criterion")
        else: # if 1d list
            naughty = list(set([x for x in param if len(x) != 10]))
            if len(naughty) > 0:
                raise ValueError(f"ValueError: Invalid '{name}' length detected, must equal 10 digits and {naughty} do not meet this criterion")

def checkDRGOverlap(drg_code, diags, procs):
    """
    Warning if DRG and procedures and/or diagnoses are present
    """
    if (drg_code != None) & (diags != None) & (procs != None):
        warnings_list = ["ValueWarning: DRG, diagnoses and procedures specified *** Diagnostic Related Groupings (DRGs) are derived from of procedure and diagnosis codes, query will return all fields relating to DRG 'OR' diagnosis and procedures."]
    elif (drg_code != None) & (procs != None):
        warnings_list = ["ValueWarning: DRG procedures specified *** Diagnostic Related Groupings (DRGs) are derived from of procedure and diagnosis codes, query will return all fields relating to DRG 'OR' procedures."]
    elif (drg_code != None) & (procs != None):
        warnings_list = ["ValueWarning: DRG and diagnoses specified *** Diagnostic Related Groupings (DRGs) are derived from of procedure and diagnosis codes, query will return all fields relating to DRG 'OR' diagnosis."]
    else:
        warnings_list = []

    return warnings_list

def checkNPIOverlap(taxonomies, ref_npi, npi):
    """
    Warning if NPI fields and taxonomies are present
    """
    if (taxonomies != None) & (ref_npi != None) & (npi != None):
        warnings_list = ["ValueWarning: Provider Specialty/Taxonomy, Referring NPI and Primary NPI Specified *** query will return all values that meet ALL criterion. For EITHER/OR criterion, set greedy = True"]
    elif (taxonomies != None) & (ref_npi != None):
        warnings_list = ["ValueWarning: Provider Specialty/Taxonomy and Referring NPI *** query will return all values that meet BOTH criterion. For EITHER/OR criterion, set greedy = True"]
    elif (taxonomies != None) & (npi != None):
        warnings_list = ["ValueWarning: Provider Specialty/Taxonomy and NPI *** query will return all values that meet BOTH criterion. For EITHER/OR criterion, set greedy = True"]
    else:
        warnings_list = []

    if (ref_npi != None) & (npi != None):
        warnings_list.append("ValueWarning: Referring NPI and Primary NPI Specified *** query will return all values that meet BOTH criterion. For EITHER/OR criterion, set greedy = True")
    
    return warnings_list

def checkLevels(param, name, levels = None):
    """
    Warning if list dimensionality is off
    """
    if (param != None) & (levels != None):
        if type(param) == list:
            if (any(isinstance(el, list) for el in param)) & (levels == 1): # if list of list
                raise ValueError(f"ValueError: List of lists implies an 'and' statement and is not allowed for '{name}' in this type of query")
            elif (any(isinstance(el, list) for el in param)) & (levels > 2):
                raise ValueError(f"ValueError: List of lists implies an 'and' statement and is not allowed beyond two lists for '{name}' in this type of query")

def checkFieldValues(param, field_values, name):
    """
    Check if param is a valid field value
    """
    if param != None:

        try:
            values = list(field_values[name].keys()) # nested dicts
        except:
            values = list(field_values[name]) # basic lists

        if (type(param) == str) & (param not in values):
            raise ValueError(f"ValueError: '{param}' is not a valid '{name}' value, please reference self.field_values.list_values(field = '{name}') for full list of options")
        elif type(param) == list:
            if any(isinstance(el, list) for el in param): # if list of list
                naughty = list(set([item for sublist in param for item in sublist if item not in values]))
                if len(naughty) > 0:
                    raise ValueError(f"ValueError: {param} are not a valid '{name}' value(s), please reference self.field_values.list_values(field = '{name}') for full list of options")
            else: # if 1d list
                naughty = list(set([x for x in param if x not in values]))
                if len(naughty) > 0:
                    raise ValueError(f"ValueError: {param} are not a valid '{name}' value(s), please reference self.field_values.list_values(field = '{name}') for full list of options")

def checkDRGConflate(drg_code, proc_code, proc_subcategory, proc_category, diag_code, diag_subcategory, diag_category, short_diag_code):
            
            if (drg_code != None) & ((proc_code != None) | (proc_subcategory != None) | (proc_category != None)):
                raise ValueError("DRG Conflation Error: DRGs cannot be specified alongside any procedure parameters because DRGs use procedures within thier computation, conflating query the output.")
            if (drg_code != None) & ((diag_code != None) | (diag_subcategory != None) | (diag_category != None) | (short_diag_code != None)):
                raise ValueError("DRG Conflation Error: DRGs cannot be specified alongside any diagnosis parameters because DRGs use diagnoses within thier computation, conflating query the output.")

### Normalization Helpers ### 

# wrap strings in list if needed
def listWrapString(x):
    if type(x) == str:
        x = [x]
    return x

# helper string type class
def stringNorm(param, stringtype = None):
    """
    Normalize string to case, whether alone, in list or list of lists
    """
    if (stringtype == None) | (stringtype == "lower"):
        if (type(param) == int) | (type(param) == str):
            param = str(param).lower().strip()
        elif type(param) == list:
            if any(isinstance(el, list) for el in param): # if list of list
                param = [[str(x).lower().strip() for x in y] if type(y) == list else [str(y).lower().strip()] for y in param]
            else: # 1d list
                param = [str(x).lower().strip() for x in param]
    elif stringtype == "upper":
        if (type(param) == int) | (type(param) == str):
            param = str(param).upper().strip()
        elif type(param) == list:
            if any(isinstance(el, list) for el in param): # if list of list
                param = [[str(x).upper().strip() for x in y] if type(y) == list else [str(y).upper().strip()] for y in param]
            else: # 1d list
                param = [str(x).upper().strip() for x in param]
    elif stringtype == "title":
        if (type(param) == int) | (type(param) == str):
            param = str(param).title().strip()
        elif type(param) == list:
            if any(isinstance(el, list) for el in param): # if list of list
                param = [[str(x).title().strip() for x in y] if type(y) == list else [str(y).title().strip()] for y in param]
            else: # 1d list
                param = [str(x).title().strip() for x in param]

    return param

# helper string code norm
def codeNorm(code, n = None, punct = None):
    """
    Normalize code to final format
    """
    # remove punctuation
    if (code != None) & (punct == False):
        if (type(code) == int) | (type(code) == str):
            code = str(re.sub('[^A-Za-z0-9]+', '', code))
        elif type(code) == list:
            if any(isinstance(el, list) for el in code): # if list of list
                code = [[re.sub('[^A-Za-z0-9]+', '', str(x)) for x in y] if type(y) == list else [re.sub('[^A-Za-z0-9]+', '', str(y))] for y in code]
            else: # if 1d list
                code = [re.sub('[^A-Za-z0-9]+', '', str(x)) for x in code]
    
    # remove punctuation
    if (code != None) & (n != None):
        if type(code) == int:
            code = str(code)
        if (type(code) == str) & (len(code) < n): # add when less
            code = code.rjust(n, '0')
        if (type(code) == str) & (len(code) > n): # cut when more
            code = code[:n]
        elif type(code) == list:
            if any(isinstance(el, list) for el in code): # if list of list
                code = [[str(x).rjust(n,'0')[:n] for x in y] if type(y) == list else [str(y).rjust(n,'0')[:n]] for y in code]
            else: # if 1d list
                code = [str(x).rjust(n,'0')[:n] for x in code]

    return code

# helper quality and npi normalization function
def NPINorm(param):
    """
    Normalize NPI values
    """
    # normalize input
    if type(param) == int: # int to str
        param = str(param).replace(".0","").strip()
    elif type(param) == list: # if a list
        if any(isinstance(el, list) for el in param): # if list of list
            param = [[str(x).replace(".0","").strip() for x in y] if type(y) == list else [str(y).replace(".0","").strip()] for y in param]
        else: # if 1d list
            param = [str(x).replace(".0","").strip() for x in param]
            
    return param

### Logical Prioritization ###

def placeOfServicePriority(place_of_service, field_values):
    """
    Extract underlying values from place of service
    """
    if place_of_service != None:
        # extract, convert to list
        if (type(place_of_service) == str) | (type(place_of_service) == int):
            place_of_service = [str(place_of_service)]
        
        # convert names to codes in lists
        if (type(place_of_service) == list) & (len(place_of_service) > 0):
            if (any(isinstance(el, list) for el in place_of_service)):
                pos = [[field_values["place_of_service"][x] for x in y] for y in place_of_service] # list of list logic 
            else:
                pos = [field_values["place_of_service"][x] for x in place_of_service] # list logic

    else:
        pos = None

    return pos

def diagnosisPriority(diag_code, short_diag_code, diag_subcategory, diag_category, diag_greedy, field_values):
    """
    Surface underlying short diagnosis codes for text and string according to diagnosis greedy policy
    """
    # control for all nones
    if (diag_code == None) & (short_diag_code == None) & (diag_subcategory == None) & (diag_category == None):
        diags = None
        warnings_list = []
    else:
        # determine if we need to be greedy at any point within diagnoses
        if (diag_code != None):
            next
            if (any(isinstance(el, list) for el in diag_code)):
                diag_greedy = False
        if (short_diag_code != None):
            next
            if (any(isinstance(el, list) for el in short_diag_code)):
                diag_greedy = False
        if (diag_subcategory != None):
            next
            if (any(isinstance(el, list) for el in diag_subcategory)):
                diag_greedy = False
        if (diag_category != None):
            next
            if (any(isinstance(el, list) for el in diag_category)):
                diag_greedy = False
        
        # set default greediness
        if diag_greedy != False:
            diag_greedy = True

        # proper warnings
        if diag_greedy == False:
            warnings_item2 = "Non-Greedy Diagnosis Priority *** this query will use 'AND' logic for diagnostic list selections, all diagnoses within sub-lists MUST be present in order to be returned."
        else:
            warnings_item2 = "Greedy Diagnosis Priority *** this query will use 'OR' logic for diagnostic selections, meaning that query will return all results where ANY diagnosis is present."

        # multiple category warning
        if len([x for x in [diag_code, short_diag_code, diag_subcategory, diag_category] if x != None]) > 1:
            warnings_item1 = f"Multiple Diagnosis Options Selected: {warnings_item2}"
            warnings_list = [warnings_item1]
        else:
            warnings_list = []

        # instantiate diags
        diags = []

        if diag_code != None:
            # if string or int, make list
            if (type(diag_code) == str) | (type(diag_code) == int):
                diag_code = [str(diag_code)]
            # if not greedy and not nested, make it so
            if (diag_greedy == False) &  (any(isinstance(el, list) for el in diag_code) == False):
                diag_code = [diag_code]
            # if list and no length, make diags
            if (type(diag_code) == list) & (len(diag_code) != 0):
                diags.extend(diag_code)

        if short_diag_code != None:
            # if string or int, make list
            if (type(short_diag_code) == str) | (type(short_diag_code) == int):
                short_diag_code = [str(short_diag_code)]
            # if not greedy and not nested, make it so
            if (diag_greedy == False) &  (any(isinstance(el, list) for el in short_diag_code) == False):
                short_diag_code = [short_diag_code]
            # if list and no length, make diags
            if (type(short_diag_code) == list) & (len(short_diag_code) != 0):
                diags.extend(short_diag_code)

        if diag_subcategory != None:

            # string or int, make list
            if (type(diag_subcategory) == str) | (type(diag_subcategory) == int):
                diag_subcategory = [str(diag_subcategory)]
            # if not greedy and not nested, make it so
            if (diag_greedy == False) &  (any(isinstance(el, list) for el in diag_subcategory) == False):
                diag_subcategory = [diag_subcategory]
            # list and no length, make diags the list specified
            if (type(diag_subcategory) == list) & (len(diags) == 0):
                diags = [x[1] for x in field_values["diag_subcategory"].items() if x[0] in diag_subcategory]
                if diag_greedy == True:
                    try:
                        diags = [item for sublist in diags for item in sublist] # unnest
                    except:
                        next

            # nested list (default non-greedy), bring up underlying codes & append list
            elif any(isinstance(el, list) for el in diag_subcategory): 
                temp2 = []
                for i in range(0,len(diag_subcategory)):
                    temp = [x[1] for x in field_values["diag_subcategory"].items() if x[0] in diag_subcategory[i]]
                    temp2.extend(temp)
                diags.extend(temp2)

            # non-greedy list, , bring up underlying codes & extend list wrapped values
            elif (type(diag_subcategory) == list) & (diag_greedy == False):
                temp = [x[1] for x in field_values["diag_subcategory"].items() if x[0] in diag_subcategory]
                try:
                    temp = [item for sublist in temp for item in sublist] # unnest
                except:
                    next
                diags.extend([temp])
                
            # greedy list, bring up underlying codes & extend values
            elif (type(diag_subcategory) == list) & (diag_greedy == True):
                temp = [x[1] for x in field_values["diag_subcategory"].items() if x[0] in diag_subcategory]
                try:
                    temp = [item for sublist in temp for item in sublist] # unnest
                except:
                    next
                diags.extend(temp)

        if diag_category != None:

            # string or int, make list
            if (type(diag_category) == str) | (type(diag_category) == int):
                diag_category = [str(diag_category)]
            # if not greedy and not nested, make it so
            if (diag_greedy == False) &  (any(isinstance(el, list) for el in diag_category) == False):
                diag_category = [diag_category]
            # list and no length, make diags the list specified
            if (type(diag_category) == list) & (len(diags) == 0):
                diags = [x[1] for x in field_values["diag_category"].items() if x[0] in diag_category]
                if diag_greedy == True:
                    try:
                        diags = [item for sublist in diags for item in sublist] # unnest
                    except:
                        next

            # nested list (default non-greedy), bring up underlying codes & append list
            elif any(isinstance(el, list) for el in diag_category): 
                temp2 = []
                for i in range(0,len(diag_category)):
                    temp = [x[1] for x in field_values["diag_category"].items() if x[0] in diag_category[i]]
                    temp2.extend(temp)
                diags.extend(temp2)

            # non-greedy list, , bring up underlying codes & extend list wrapped values
            elif (type(diag_category) == list) & (diag_greedy == False):
                temp = [x[1] for x in field_values["diag_category"].items() if x[0] in diag_category]
                try:
                    temp = [item for sublist in temp for item in sublist] # unnest
                except:
                    next
                diags.extend([temp])
                
            # greedy list, bring up underlying codes & extend values
            elif (type(diag_category) == list) & (diag_greedy == True):
                temp = [x[1] for x in field_values["diag_category"].items() if x[0] in diag_category]
                try:
                    temp = [item for sublist in temp for item in sublist] # unnest
                except:
                    next
                diags.extend(temp)

        # drop any empty lists
        diags = [x for x in diags if len(x) > 0]

        # ensure none if empty
        if len(diags) == 0:
            diags = None

        # wrap any loose items if list of lists
        if any(isinstance(el, list) for el in diags):
            diags = [[x] if type(x) != list else x for x in diags]

    return diags, diag_greedy, warnings_list

def procedurePriority(proc_code, proc_subcategory, proc_category, proc_greedy, field_values, query_type = None):
    """
    Surface underlying procedure codes for text and string according to procedure greedy policy
    """
    # set default to journey, alternative is supplemental
    if query_type == None:
        query_type = 'journey'
    
    # control for all nones
    if (proc_code == None) & (proc_subcategory == None) & (proc_category == None):
        procs = None
        warnings_list = []
    else:
        # determine if we need to be greedy at any point within procedure
        if (proc_code != None):
            next
            if (any(isinstance(el, list) for el in proc_code)):
                proc_greedy = False
        if (proc_subcategory != None):
            next
            if (any(isinstance(el, list) for el in proc_subcategory)):
                proc_greedy = False
        if (proc_category != None):
            next
            if (any(isinstance(el, list) for el in proc_category)):
                proc_greedy = False
        
        # set default greediness
        if proc_greedy != False:
            proc_greedy = True

        warnings_list = []
        # proper warnings
        if proc_greedy == False:
            warnings_item2 = "Non-Greedy Procedure Priority *** this query will use 'AND' logic for procedural list selections, all procedures within sub-lists MUST be present in order to be returned."
        else:
            warnings_item2 = "Greedy Procedure Priority *** this query will use 'OR' logic for procedural selections, meaning that query will return all results where ANY procedure is present."

        # multiple category warning
        if len([x for x in [proc_code, proc_subcategory, proc_category] if x != None]) > 1:
            warnings_item1 = f"Multiple Procedure Options Selected: {warnings_item2}"
            warnings_list = [warnings_item1]
        else:
            warnings_list = [warnings_item2]

        # instantiate procs
        procs = []

        if proc_code != None:
            # make any string or int, make list
            if (type(proc_code) == str) | (type(proc_code) == int):
                proc_code = [str(proc_code)]
            # if not greedy and not nested, make it so
            if (proc_greedy == False) & (any(isinstance(el, list) for el in proc_code) == False):
                proc_code = [proc_code]
            # if list and no length, make procs
            if (type(proc_code) == list) & (len(proc_code) > 0) & (proc_greedy == True):
                procs.extend(proc_code)
            # non-greedy nested list, normal append
            elif (proc_greedy == False) & (any(isinstance(el, list) for el in proc_code)):
                procs.extend(proc_code)
            
        if proc_subcategory != None:

            if query_type == 'journey':

                # make any string or int a list
                if (type(proc_subcategory) == str) | (type(proc_subcategory) == int):
                    proc_subcategory = [str(proc_subcategory)]
                # if not greedy and not nested, make it so
                if (proc_greedy == False) &  (any(isinstance(el, list) for el in proc_subcategory) == False):
                    proc_subcategory = [proc_subcategory] #TODO: Need to fix this to get the proc codes
                # if nested list, underlying code matching doesn't matter bc values are exclusive from one another
                if (len(procs) == 0) & (any(isinstance(el, list) for el in proc_subcategory)):
                    procs.extend(proc_subcategory) #TODO: Need to check that this is working - as right now it just appends the subcategory names to the list
                # if greedy list (most common), remove overlapping codes
                elif (type(proc_subcategory) == list):
                    procs.extend(proc_subcategory)

            elif query_type == "supplemental":

                # make any string or int a list
                if (type(proc_subcategory) == str) | (type(proc_subcategory) == int):
                    proc_subcategory = [str(proc_subcategory)]
                # if not greedy and not nested, make it so
                if (proc_greedy == False) &  (any(isinstance(el, list) for el in proc_subcategory) == False):
                    proc_subcategory = [proc_subcategory] #TODO: Need to fix this to get the proc codes
                # if nested list, underlying code matching doesn't matter bc values are exclusive from one another
                if (len(procs) == 0) & (any(isinstance(el, list) for el in proc_subcategory)):
                    procs.extend(proc_subcategory) #TODO: Need to check that this is working - as right now it just appends the subcategory names to the list
                # if greedy list (most common), remove overlapping codes
                elif (type(proc_subcategory) == list):
                    try:
                        underlying_procs = [x[1] for x in field_values["proc_subcategory"].items() if x[0] in proc_subcategory] # identify any codes overlapping categories from procs
                        underlying_procs = [item for sublist in underlying_procs for item in sublist] # flatten
                        procs = [x for x in procs if x not in underlying_procs] # remove overlapping procs
                    except:
                        next
                    procs.extend(underlying_procs)
             
        if proc_category != None:

            if query_type == 'journey':

                # make any string or int a list
                if (type(proc_category) == str) | (type(proc_category) == int):
                    proc_category = [str(proc_category)]
                # if not greedy and not nested, make it so
                if (proc_greedy == False) &  (any(isinstance(el, list) for el in proc_category) == False):
                    proc_category = [proc_category]
                # if nested list or not greedy, underlying code matching doesn't matter bc values are exclusive from one another
                if (len(procs) == 0) | (any(isinstance(el, list) for el in proc_category)):
                    if (any(isinstance(el, list) for el in proc_category)):
                        # nested list 
                        temp = []
                        for cat in proc_category:
                            temp0 = [x for x in list(field_values["proc_category"].keys()) if x.split(" - ")[0] in cat] # isolate subcategory list of categories for common vocab
                            temp.extend(temp0)
                        procs.append(temp)
                    else:
                        # new list
                        procs = [x[1] for x in list(field_values["proc_category"].keys()) if x.split(" - ")[0] in proc_category] # isolate subcategory list of categories for common vocab
                        procs = []
                        for cat in proc_category:
                            #Get the list of procs associated with the category
                            dic = field_values['proc_category']
                            temp_procs = dic[cat]
                            procs.extend(temp_procs)

                # greedy list (most common), remove overlapping codes
                elif (type(proc_category) == list):
                    temp0 = [x for x in list(field_values["proc_subcategory"].keys()) if x.split(" - ")[0] in proc_category] # isolate subcategory list of categories for common vocab
                    temp0 = [x for x in temp0 if x not in procs] # remove any overlapping subcategories
                    try:
                        underlying_procs = [x[1] for x in field_values["proc_subcategory"].items() if x[0] in temp0] # identify any codes overlapping categories from procs
                        underlying_procs = [item for sublist in underlying_procs for item in sublist] # flatten
                        procs = [x for x in procs if x not in underlying_procs] # remove overlapping procs
                    except:
                        next
                    procs.extend(temp0)

            elif query_type == 'supplemental':

                # make any string or int a list
                if (type(proc_category) == str) | (type(proc_category) == int):
                    proc_category = [str(proc_category)]
                # if not greedy and not nested, make it so
                if (proc_greedy == False) &  (any(isinstance(el, list) for el in proc_category) == False):
                    proc_category = [proc_category] #TODO: Need to fix this to get the proc codes
                # if nested list, underlying code matching doesn't matter bc values are exclusive from one another
                if (len(procs) == 0) & (any(isinstance(el, list) for el in proc_category)):
                    procs.extend(proc_category) #TODO: Need to check that this is working - as right now it just appends the subcategory names to the list
                # if greedy list (most common), remove overlapping codes
                elif (type(proc_category) == list):
                    try:
                        underlying_procs = [x[1] for x in field_values["proc_category"].items() if x[0] in proc_category] # identify any codes overlapping categories from procs
                        underlying_procs = [item for sublist in underlying_procs for item in sublist] # flatten
                        procs = [x for x in procs if x not in underlying_procs] # remove overlapping procs
                    except:
                        next
                    procs.extend(underlying_procs)

        # drop any empty lists
        procs = [x for x in procs if len(x) > 0]

        if len(procs) == 0:
            procs = None

        # wrap any loose items if list of lists
        if any(isinstance(el, list) for el in procs):
            procs = [[x] if type(x) != list else x for x in procs]

    return procs, proc_greedy, warnings_list

def taxonomyPriority(taxonomy_code, subspecialty, specialty, specialty_category, taxonomy_greedy, field_values):
    """
    Surface underlying taxonomy codes for text and string according to specialty/taxonomy greedy policy
    """
    # control for all nones
    if (taxonomy_code == None) & (subspecialty == None) & (specialty == None) & (specialty_category == None):
        taxonomies = None
        warnings_list = []
    else:    
        # determine if we need to be greedy at any point within taxonomyedure
        if (taxonomy_code != None):
            next
            if (any(isinstance(el, list) for el in taxonomy_code)):
                taxonomy_greedy = False
        if (subspecialty != None):
            next
            if (any(isinstance(el, list) for el in subspecialty)):
                taxonomy_greedy = False
        if (specialty != None):
            next
            if (any(isinstance(el, list) for el in specialty)):
                taxonomy_greedy = False
        if (specialty_category != None):
            next
            if (any(isinstance(el, list) for el in specialty_category)):
                taxonomy_greedy = False
        
        # set default greediness
        if taxonomy_greedy != False:
            taxonomy_greedy = True

        # proper warning
        if taxonomy_greedy == False:
            warnings_item2 = "Non-Greedy Taxonomy Priority *** this query will use 'AND' logic for taxonomy list selections, all taxonomies within sub-lists MUST be present in order to be returned."
        else:
            warnings_item2 = "Greedy Taxonomy Priority *** this query will use 'OR' logic for taxonomy selections, meaning that query will return all results where ANY taxonomy is present."

        # multiple category warning
        if len([x for x in [taxonomy_code, subspecialty, specialty, specialty_category] if x != None]) > 1:
            warnings_item1 = f"Multiple Specialty/Taxonmoy Options Selected: {warnings_item2}"
            warnings_list = [warnings_item1]
        else:
            warnings_list = []

        # instantiate taxonomies
        taxonomies = []

        if taxonomy_code != None:
            # if string or int, make list
            if (type(taxonomy_code) == str) | (type(taxonomy_code) == int):
                taxonomy_code = [str(taxonomy_code)]
            # if not greedy and not nested, make it so
            if (taxonomy_greedy == False) &  (any(isinstance(el, list) for el in taxonomy_code) == False):
                taxonomy_code = [taxonomy_code]
            # if list and no length, make taxonomies
            if (type(taxonomy_code) == list) & (len(taxonomy_code) > 0):
                taxonomies.extend(taxonomy_code)

        if subspecialty != None:

            # string or int, make list
            if (type(subspecialty) == str) | (type(subspecialty) == int):
                subspecialty = [str(subspecialty)]
            # if not greedy and not nested, make it so
            if (taxonomy_greedy == False) &  (any(isinstance(el, list) for el in subspecialty) == False):
                subspecialty = [subspecialty]

            # list and no length, make taxonomies the list specified
            if (type(subspecialty) == list) & (len(taxonomies) == 0):
                taxonomies = [x[1] for x in field_values["subspecialty"].items() if x[0] in subspecialty]
                if taxonomy_greedy == True:
                    try:
                        taxonomies = [item for sublist in taxonomies for item in sublist] # unnest
                    except:
                        next

            # nested list (default non-greedy), bring up underlying codes & append list
            elif any(isinstance(el, list) for el in subspecialty): 
                temp2 = []
                for i in range(0,len(subspecialty)):
                    temp = [x[1] for x in field_values["subspecialty"].items() if x[0] in subspecialty[i]]
                    temp2.extend(temp)
                taxonomies.extend(temp2)

            # non-greedy list, , bring up underlying codes & extend list wrapped values
            elif (type(subspecialty) == list) & (taxonomy_greedy == False):
                temp = [x[1] for x in field_values["subspecialty"].items() if x[0] in subspecialty]
                try:
                    temp = [item for sublist in temp for item in sublist] # unnest
                except:
                    next
                taxonomies.extend([temp])
                
            # greedy list, bring up underlying codes & extend values
            elif (type(subspecialty) == list) & (taxonomy_greedy == True):
                temp = [x[1] for x in field_values["subspecialty"].items() if x[0] in subspecialty]
                try:
                    temp = [item for sublist in temp for item in sublist] # unnest
                except:
                    next
                taxonomies.extend(temp)

        if specialty != None:

            # string or int, make list
            if (type(specialty) == str) | (type(specialty) == int):
                specialty = [str(specialty)]
            # if not greedy and not nested, make it so
            if (taxonomy_greedy == False) &  (any(isinstance(el, list) for el in specialty) == False):
                specialty = [specialty]

            # list and no length, make taxonomies the list specified
            if (type(specialty) == list) & (len(taxonomies) == 0):
                taxonomies = [x[1] for x in field_values["specialty"].items() if x[0] in specialty]
                if taxonomy_greedy == True:
                    try:
                        taxonomies = [item for sublist in taxonomies for item in sublist] # unnest
                    except:
                        next

            # nested list (default non-greedy), bring up underlying codes & append list
            elif any(isinstance(el, list) for el in specialty): 
                temp2 = []
                for i in range(0,len(specialty)):
                    temp = [x[1] for x in field_values["specialty"].items() if x[0] in specialty[i]]
                    temp2.extend(temp)
                taxonomies.extend(temp2)

            # non-greedy list, , bring up underlying codes & extend list wrapped values
            elif (type(specialty) == list) & (taxonomy_greedy == False):
                temp = [x[1] for x in field_values["specialty"].items() if x[0] in specialty]
                try:
                    temp = [item for sublist in temp for item in sublist] # unnest
                except:
                    next
                taxonomies.extend([temp])
                
            # greedy list, bring up underlying codes & extend values
            elif (type(specialty) == list) & (taxonomy_greedy == True):
                temp = [x[1] for x in field_values["specialty"].items() if x[0] in specialty]
                try:
                    temp = [item for sublist in temp for item in sublist] # unnest
                except:
                    next
                taxonomies.extend(temp)

        if specialty_category != None:

            # string or int, make list
            if (type(specialty_category) == str) | (type(specialty_category) == int):
                specialty_category = [str(specialty_category)]
            # if not greedy and not nested, make it so
            if (taxonomy_greedy == False) &  (any(isinstance(el, list) for el in specialty_category) == False):
                specialty_category = [specialty_category]

            # list and no length, make taxonomies the list specified
            if (type(specialty_category) == list) & (len(taxonomies) == 0):
                taxonomies = [x[1] for x in field_values["specialty_category"].items() if x[0] in specialty_category]
                if taxonomy_greedy == True:
                    try:
                        taxonomies = [item for sublist in taxonomies for item in sublist] # unnest
                    except:
                        next

            # nested list (default non-greedy), bring up underlying codes & append list
            elif any(isinstance(el, list) for el in specialty_category): 
                temp2 = []
                for i in range(0,len(specialty_category)):
                    temp = [x[1] for x in field_values["specialty_category"].items() if x[0] in specialty_category[i]]
                    temp2.extend(temp)
                taxonomies.extend(temp2)

            # non-greedy list, bring up underlying codes & extend list wrapped values
            elif (type(specialty_category) == list) & (taxonomy_greedy == False):
                temp = [x[1] for x in field_values["specialty_category"].items() if x[0] in specialty_category]
                try:
                    temp = [item for sublist in temp for item in sublist] # unnest
                except:
                    next
                taxonomies.extend([temp])
                
            # greedy list, bring up underlying codes & extend values
            elif (type(specialty_category) == list) & (taxonomy_greedy == True):
                temp = [x[1] for x in field_values["specialty_category"].items() if x[0] in specialty_category]
                try:
                    temp = [item for sublist in temp for item in sublist] # unnest
                except:
                    next
                taxonomies.extend(temp)

        # drop any empty lists
        taxonomies = [x for x in taxonomies if len(x) > 0]

        if len(taxonomies) == 0:
            taxonomies = None

        # wrap any loose items if list of lists
        if any(isinstance(el, list) for el in taxonomies):
            taxonomies = [[x] if type(x) != list else x for x in taxonomies]

    return taxonomies, taxonomy_greedy, warnings_list

def geoPriority(state, division, region, metro, short_zip, field_values):
    """
    Surface underlying zip and state values for text
    """
    # control for all nones
    if (state == None) & (division == None) & (region == None) & (metro == None) & (short_zip == None):
        states2 = None
        zips = None 
        warnings_list = []

    else:

        # warning and flattening of any multi-level geo lists
        warnings_list = []
        if (state != None):
            next
            if (any(isinstance(el, list) for el in state)):
                state = [item for sublist in state for item in sublist]
                warnings_list.append("*** Multi-Level lists are not allowed within geographical fields *** query will include results for ANY 'state' fields listed")
        if (division != None):
            next
            if (any(isinstance(el, list) for el in division)):
                division = [item for sublist in division for item in sublist]
                warnings_list.append("*** Multi-Level lists are not allowed within geographical fields *** query will include results for ANY 'division' fields listed")
        if (region != None):
            next
            if (any(isinstance(el, list) for el in region)):
                region = [item for sublist in region for item in sublist]
                warnings_list.append("*** Multi-Level lists are not allowed within geographical fields *** query will include results for ANY 'region' fields listed")
        if (metro != None):
            next
            if (any(isinstance(el, list) for el in metro)):
                metro = [item for sublist in metro for item in sublist]
                warnings_list.append("*** Multi-Level lists are not allowed within geographical fields *** query will collapse list to include results for ANY 'metro' fields listed")
        if (short_zip != None):
            next
            if (any(isinstance(el, list) for el in short_zip)):
                short_zip = [item for sublist in short_zip for item in sublist]
                warnings_list.append("*** Multi-Level lists are not allowed within geographical fields *** query will collapse list to include results for ANY 'metro' fields listed")

        ### state-mapped sources ###

        # state logic
        states = []
        if state != None:
            if (type(state) == str):
                state = [state]
            if (type(state) == list) & (len(state) != 0):
                state = [x.upper() for x in state]
                states.extend(state)

        # division logic
        divisions = []
        if division != None:
            if (type(division) == str):
                division = [division]
            if (type(division) == list) & (len(division) != 0):
                temp = [x[1] for x in field_values["division"].items() if x[0] in division]
                try:
                    temp = [item for sublist in temp for item in sublist]
                except:
                    next
                divisions.extend(temp)

        # _region logic
        regions = []
        if region != None:
            if (type(region) == str):
                region = [region]
            if (type(region) == list) & (len(region) != 0):
                temp = [x[1] for x in field_values["region"].items() if x[0] in region]
                try:
                    temp = [item for sublist in temp for item in sublist]
                except:
                    next
                regions.extend(temp)

        ### zip-mapped sources ###

        # short_zip logic
        short_zips = []
        if short_zip != None:
            if (type(short_zip) == str):
                short_zip = [short_zip]
            if (type(short_zip) == list) & (len(short_zip) != 0):
                short_zips.extend(short_zip)

        # metro logic
        metros = []
        if metro != None:
            if (type(metro) == str):
                metro = [metro]
            if (type(metro) == list) & (len(metro) != 0):
                temp = [x[1] for x in field_values["metro"].items() if x[0] in metro]
                try:
                    temp = [item for sublist in temp for item in sublist]
                except:
                    next
                metros.extend(temp)

        ### Hierarchical Inclusivity Measures (in, out, both) ### 

        # intra zip priority
        if (len(short_zips) != 0) & (len(metros) != 0):
            zips = short_zips + metros
        elif (len(short_zips) != 0) & (len(metros) == 0):
            zips = short_zips
        elif (len(short_zips) == 0) & (len(metros) != 0):
            zips = metros
        else:
            zips = None

        # intra state priority
        if (len(states) != 0) & (len(divisions) != 0) & (len(regions) != 0): # all non-null
            states2 = states + divisions + regions
        elif (len(states) == 0) & (len(divisions) != 0) & (len(regions) != 0): # states non-null
            states2 = divisions + regions
        elif (len(states) != 0) & (len(divisions) != 0) & (len(regions) == 0): # regions non-null
            states2 = states + divisions
        elif (len(states) != 0) & (len(divisions) == 0) & (len(regions) != 0): # divisions non-null
            states2 = states + regions
        elif (len(states) != 0) & (len(divisions) == 0) & (len(regions) == 0): # only states
            states2 = states
        elif (len(states) == 0) & (len(divisions) == 0) & (len(regions) != 0): # only regions
            states2 = regions
        elif (len(states) == 0) & (len(divisions) != 0) & (len(regions) == 0): # only divisions
            states2 = divisions
        else:
             states2 = None

        # remove zips when overlapped with state
        if (zips != None) & (states2 != None):
            state_zips = [x[1] for x in field_values["state"].items() if x[0] in states2]
            if (any(isinstance(el, list) for el in state_zips)): 
                state_zips = [y for x in state_zips for y in x]
            overlapping_zips = [x for x in zips if x in state_zips]
            if len(overlapping_zips) > 0:
                warnings_list.append(f"Overlapping state/short_zip Selection: The following short_zip codes ({overlapping_zips}) appear within State selection(s) made, query will return state and non-overlappning short_zip results. To return excluded short_zip codes only, specify without overlapping state.")
            zips = [x for x in zips if x not in state_zips]

        ### Warnings On Multi-Selection ### 
        try:
            if (len(states) != 0) & (len(divisions) != 0) | (len(regions) != 0) & (len(divisions) != 0) | (len(states) != 0) & (len(regions) != 0) | (len(metros) != 0) & (len(zips) != 0):
                warnings_list.append("Multiple Geography Options Selected: Greedy Diagnosis Priority *** Query will return all results that corresponding to ANY of the States and/or Short Zip Codes selected.")
            elif (len(states) > 0) & (len(zips) > 0):
                warnings_list.append("Multiple Geography Options Selected: Greedy Diagnosis Priority *** Query will return all results that corresponding to ANY of the States and/or Short Zip Codes selected.")
        except:
            next
        
        if states2 != None:
            if (len(states2) == 0):
                states2 = None
        if zips != None:
            if len(zips) == 0:
                zips = None

    return states2, zips, warnings_list

def scriptPriority(ndc, substance, pharm_class, marketing_category, script_greedy, field_values):
    
    """
    Surface underlying NDCs for text and string according to script greedy policy
    """
    
    # control for all nones
    if (ndc == None) & (substance == None) & (marketing_category == None) & (pharm_class == None):
        ndcs = None
        warnings_list = []
    else:    
        # determine if we need to be greedy at any point within scripts
        if (ndc != None):
            next
            if (any(isinstance(el, list) for el in ndc)):
                script_greedy = False
        if (substance != None):
            next
            if (any(isinstance(el, list) for el in substance)):
                script_greedy = False
        if (pharm_class != None):
            next
            if (any(isinstance(el, list) for el in pharm_class)):
                script_greedy = False
        if (marketing_category != None):
            next
            if (any(isinstance(el, list) for el in marketing_category)):
                script_greedy = False
        
        # set default greediness
        if script_greedy != False:
            script_greedy = True

        # proper warning
        if script_greedy == False:
            warnings_item2 = "Non-Greedy Script Priority *** this query will use 'AND' logic for script list selections, all NDCs within sub-lists MUST be present in order to be returned."
        else:
            warnings_item2 = "Greedy Script Priority *** this query will use 'OR' logic for script selections, meaning that query will return all results where ANY NDC is present."

        # multiple category warning
        if len([x for x in [ndc, substance, pharm_class, marketing_category] if x != None]) > 1:
            warnings_item1 = f"Multiple NDC Options Selected: {warnings_item2}"
            warnings_list = [warnings_item1]
        else:
            warnings_list = []

        # instantiate ndc
        ndcs = []

        if ndc != None:
            # if string or int, make list
            if (type(ndc) == str) | (type(ndc) == int):
                ndc = [str(ndc)]
            # if not greedy and not nested, make it so
            if (script_greedy == False) &  (any(isinstance(el, list) for el in ndc) == False):
                ndc = [ndc]
            # if list and no length, make ndcs
            if (type(ndc) == list) & (len(ndc) > 0):
                if any(isinstance(el, list) for el in ndc) == True:
                    ndc = [[ndcToNdc11(item) for item in sublist] for sublist in ndc] # nested list ndc to ndc 11
                else:
                    ndc = [ndcToNdc11(x) for x in ndc] # list ndc to ndc 11
                ndcs.extend(ndc)

        if substance != None:

            # string or int, make list
            if (type(substance) == str) | (type(substance) == int):
                substance = [str(substance)]
            # if not greedy and not nested, make it so
            if (script_greedy == False) &  (any(isinstance(el, list) for el in substance) == False):
                substance = [substance]

            # list and no length, make ndcs the list specified
            if (type(substance) == list) & (len(ndcs) == 0):
                ndcs = [x[1] for x in field_values["substance"].items() if x[0] in substance]
                if script_greedy == True:
                    try:
                        ndcs = [item for sublist in ndcs for item in sublist] # unnest
                    except:
                        next

            # nested list (default non-greedy), bring up underlying codes & append list
            elif any(isinstance(el, list) for el in substance): 
                temp2 = []
                for i in range(0,len(substance)):
                    temp = [x[1] for x in field_values["substance"].items() if x[0] in substance[i]]
                    temp2.extend(temp)
                ndcs.extend(temp2)

            # non-greedy list, , bring up underlying codes & extend list wrapped values
            elif (type(substance) == list) & (script_greedy == False):
                temp = [x[1] for x in field_values["substance"].items() if x[0] in substance]
                try:
                    temp = [item for sublist in temp for item in sublist] # unnest
                except:
                    next
                ndcs.extend([temp])
                
            # greedy list, bring up underlying codes & extend values
            elif (type(substance) == list) & (script_greedy == True):
                temp = [x[1] for x in field_values["substance"].items() if x[0] in substance]
                try:
                    temp = [item for sublist in temp for item in sublist] # unnest
                except:
                    next
                ndcs.extend(temp)

        if pharm_class != None:

            # string or int, make list
            if (type(pharm_class) == str) | (type(pharm_class) == int):
                pharm_class = [str(pharm_class)]
            # if not greedy and not nested, make it so
            if (script_greedy == False) &  (any(isinstance(el, list) for el in pharm_class) == False):
                pharm_class = [pharm_class]

            # list and no length, make ndcs the list specified
            if (type(pharm_class) == list) & (len(ndcs) == 0):
                ndcs = [x[1] for x in field_values["pharm_class"].items() if x[0] in pharm_class]
                if script_greedy == True:
                    try:
                        ndcs = [item for sublist in ndcs for item in sublist] # unnest
                    except:
                        next

            # nested list (default non-greedy), bring up underlying codes & append list
            elif any(isinstance(el, list) for el in pharm_class): 
                temp2 = []
                for i in range(0,len(pharm_class)):
                    temp = [x[1] for x in field_values["pharm_class"].items() if x[0] in pharm_class[i]]
                    temp2.extend(temp)
                ndcs.extend(temp2)

            # non-greedy list, , bring up underlying codes & extend list wrapped values
            elif (type(pharm_class) == list) & (script_greedy == False):
                temp = [x[1] for x in field_values["pharm_class"].items() if x[0] in pharm_class]
                try:
                    temp = [item for sublist in temp for item in sublist] # unnest
                except:
                    next
                ndcs.extend([temp])
                
            # greedy list, bring up underlying codes & extend values
            elif (type(pharm_class) == list) & (script_greedy == True):
                temp = [x[1] for x in field_values["pharm_class"].items() if x[0] in pharm_class]
                try:
                    temp = [item for sublist in temp for item in sublist] # unnest
                except:
                    next
                ndcs.extend(temp)

        if marketing_category != None:

            # string or int, make list
            if (type(marketing_category) == str) | (type(marketing_category) == int):
                marketing_category = [str(marketing_category)]
            # if not greedy and not nested, make it so
            if (script_greedy == False) &  (any(isinstance(el, list) for el in marketing_category) == False):
                marketing_category = [marketing_category]

            # list and no length, make ndcs the list specified
            if (type(marketing_category) == list) & (len(ndcs) == 0):
                ndcs = [x[1] for x in field_values["marketing_category"].items() if x[0] in marketing_category]
                if script_greedy == True:
                    try:
                        ndcs = [item for sublist in ndcs for item in sublist] # unnest
                    except:
                        next

            # nested list (default non-greedy), bring up underlying codes & append list
            elif any(isinstance(el, list) for el in marketing_category): 
                temp2 = []
                for i in range(0,len(marketing_category)):
                    temp = [x[1] for x in field_values["marketing_category"].items() if x[0] in marketing_category[i]]
                    temp2.extend(temp)
                ndcs.extend(temp2)

            # non-greedy list, bring up underlying codes & extend list wrapped values
            elif (type(marketing_category) == list) & (script_greedy == False):
                temp = [x[1] for x in field_values["marketing_category"].items() if x[0] in marketing_category]
                try:
                    temp = [item for sublist in temp for item in sublist] # unnest
                except:
                    next
                ndcs.extend([temp])
                
            # greedy list, bring up underlying codes & extend values
            elif (type(marketing_category) == list) & (script_greedy == True):
                temp = [x[1] for x in field_values["marketing_category"].items() if x[0] in marketing_category]
                try:
                    temp = [item for sublist in temp for item in sublist] # unnest
                except:
                    next
                ndcs.extend(temp)

        # drop any empty lists
        ndcs = [x for x in ndcs if len(x) > 0]

        if len(ndcs) == 0:
            ndcs = None

        # wrap any loose items if list of lists
        if any(isinstance(el, list) for el in ndcs):
            ndcs = [[x] if type(x) != list else x for x in ndcs]

    return ndcs, script_greedy, warnings_list

def drgPriority(drg_code, drg_category, drg_greedy, field_values):
    """
    Surface underlying drg codes for text and string according to drg greedy policy
    """
    # control for all nones
    if (drg_code == None) & (drg_category == None):
        drgs = None
        warnings_list = []
    else:
        
        # determine if we need to be greedy at any point within drgs
        if (drg_code != None):
            try:
                if (any(isinstance(el, list) for el in drg_code)):
                    drg_greedy = False
            except:
                next
        if (drg_category != None):
            try:
                if (any(isinstance(el, list) for el in drg_category)):
                    drg_greedy = False
            except:
                next
        
        # set default greediness
        if drg_greedy != False:
            drg_greedy = True

        # proper warning
        if drg_greedy == False:
            warnings_item2 = "Non-Greedy DRG Priority *** this query will use 'AND' logic for DRG list selections, all drgs within sub-lists MUST be present in order to be returned."
        else:
            warnings_item2 = "Greedy DRG Priority *** this query will use 'OR' logic for DRG selections, meaning that query will return all results where ANY DRG is present."

        # multiple category warning
        if len([x for x in [drg_code, drg_category] if x != None]) > 1:
            warnings_item1 = f"Multiple DRG Options Selected: {warnings_item2}"
            warnings_list = [warnings_item1]
        else:
            warnings_list = []

        # instantiate drgs
        drgs = []

        if drg_code != None:
            # if string or int, make list
            if (type(drg_code) == str) | (type(drg_code) == int):
                drg_code = [str(drg_code)]
            # if not greedy and not nested, make it so
            if (drg_greedy == False) &  (any(isinstance(el, list) for el in drg_code) == False):
                drg_code = [drg_code]
            # if list and no length, make drgs
            if (type(drg_code) == list) & (len(drg_code) > 0):
                drgs.extend(drg_code)

        if drg_category != None:

            # string or int, make list
            if (type(drg_category) == str) | (type(drg_category) == int):
                drg_category = [str(drg_category)]
            # if not greedy and not nested, make it so
            if (drg_greedy == False) &  (any(isinstance(el, list) for el in drg_category) == False):
                drg_category = [drg_category]

            # list and no length, make drgs the list specified
            if (type(drg_category) == list) & (len(drg_category) == 0):
                drgs = [x[1] for x in field_values["drg_category"].items() if x[0] in drg_category]
                if drg_greedy == True:
                    try:
                        drg_greedy = [item for sublist in drg_greedy for item in sublist] # unnest
                    except:
                        next

            # nested list (default non-greedy), bring up underlying codes & append list
            elif any(isinstance(el, list) for el in drg_category): 
                temp2 = []
                for i in range(0,len(drg_category)):
                    temp = [x[1] for x in field_values["drg_category"].items() if x[0] in drg_category[i]]
                    temp2.extend(temp)
                drgs.extend(temp2)

            # non-greedy list, , bring up underlying codes & extend list wrapped values
            elif (type(drg_category) == list) & (drg_greedy == False):
                temp = [x[1] for x in field_values["drg_category"].items() if x[0] in drg_category]
                try:
                    temp = [item for sublist in temp for item in sublist] # unnest
                except:
                    next
                drgs.extend([temp])
                
            # greedy list, bring up underlying codes & extend values
            elif (type(drg_category) == list) & (drg_greedy == True):
                temp = [x[1] for x in field_values["drg_category"].items() if x[0] in drg_category]
                try:
                    temp = [item for sublist in temp for item in sublist] # unnest
                except:
                    next
                drgs.extend(temp)

        # drop any empty lists
        try:
            if type(drgs[0]) == list:
                drgs = [x for x in drgs if len(x) > 0]
        except:
            next
        if len(drgs) == 0:
            drgs = None

        # wrap any loose items if list of lists
        if any(isinstance(el, list) for el in drgs):
            drgs = [[x] if type(x) != list else x for x in drgs]

    return drgs, drg_greedy, warnings_list



# COPYRIGHT 2022 MONOOCLE INSIGHTS
# ALL RIGHTS RESERVED
# UNAUTHORIZED USE, REPRODUCTION, OR
# DISTRIBUTION STRICTLY PROHIBITED
#
# AUTHOR: Raymond Deiotte
# CREATION DATE: <DATE>
# LAST UPDATED: <DATE>


import pandas as pd


class ThisClass:

    def __init__(self):
        '''
        Method to intialize the class
        '''

    def doIt(self):
        '''
        Method to control the class
        '''


if __name__ == '__main__':
    c = ThisClass()

    c.doIt()



# COPYRIGHT 2022 MONOOCLE INSIGHTS
# ALL RIGHTS RESERVED
# UNAUTHORIZED USE, REPRODUCTION, OR
# DISTRIBUTION STRICTLY PROHIBITED
#
# AUTHOR: Sean O'Malley and Raymond Deiotte
# CREATION DATE: 20230802
# LAST UPDATED: 20230821

### Import Packages/Dependencies ###

# basic

# custom
#FIXME Use for PyPi packaging
import care_query.care_query_utils as cqu

#FIXME Use for local testing
# import care_query_utils as cqu


### Field Value Data Class ###
class FieldValueData:

    def __init__(self, data):

        self.data = data

    def list_fields(self, contains = None):

        # list fields
        fields = list(self.data.keys())

        if contains != None:

            # normalize field string
            contains = cqu.stringNorm(contains)

            fields = [x for x in fields if contains in x]

        return fields

    def list_values(self, field, contains = None):

        # normalize field string
        field = cqu.stringNorm(field)

        try:

            # get corresponding values
            values = list(self.data[field])

            if contains != None:

                # normalize value_contains string
                contains = cqu.stringNorm(contains)

                values = [x for x in values if contains in x]

            return values
            
        except:
            raise ValueError(f"{field} is not an available field, please reference self.field_values.list_fields() for full list of options.")



# COPYRIGHT 2022 MONOOCLE INSIGHTS
# ALL RIGHTS RESERVED
# UNAUTHORIZED USE, REPRODUCTION, OR
# DISTRIBUTION STRICTLY PROHIBITED
#
# AUTHOR: Raymond Deiotte and Sean O'Malley
# CREATION DATE: 20230822
# LAST UPDATED: 20230822

### Import Packages ###
import json
import importlib.resources

### Data Dictionary ###

# create data table description 
tables = {"JOURNEY_TABLE":{"description":"This dataset offers comprehensive detail of longitudinal patient journeys, with line-item encounter detail, insight into payor, provider, procedure, revenue code, charges and SmartAllowedAmount®.", 
                           "queries":["careEncounter", "careVisit", "careEpisode", "careJourney"], 
                           "features":{"journey_id": "Randomly generated identifier unique to each longitudinal patient journey.",
                                        "episode_id": "Randomly generated identifier unique to each sequence of encounters with at least 30 days between visits.",
                                        "visit_id": "Randomly generated identifier unique to each set of encounters that occur within the same event of care.",
                                        "encounter_id": "Randomly generated identifier, unique to an isolated event of care.",
                                        "date": "Primary date of each encounter.",
                                        "patient_age": "Age (within 5 years) of the patient on each encounter.",
                                        "patient_state": "US State residence of the patient on each encounter.",
                                        "patient_zip3": "First three zip code digits of the patient on each encounter.",
                                        "place_of_service": "Type of facility where a encounter occurred.",
                                        "visit_type": "Description of the setting of care.",
                                        "payor": "Name of the primary payor on each encounter.",
                                        "payor_channel": "Payor channel-of-business, line-of-business, or plan type.",
                                        "ref_npi": "Referring individual provider of an encounter, identified via their National Provider Identifier (NPI).",
                                        "hcp_npi": "Primary individual provider of an encounter, identified via their National Provider Identifier (NPI).",
                                        "hcp_taxonomy": "Primary taxonomy code of the individual provider on an encounter.",
                                        "hcp_specialty": "Primary specialty of the individual provider provider on an encounter.",
                                        "hco_npi": "Organizational provider of an encounter, identified via their National Provider Identifier (NPI).",
                                        "diag_1": "Primary diagnosis listed on each encounter, can be represented as an ICD10 code.",
                                        "diag_2": "Secondary diagnosis listed on each encounter, can be represented as an ICD10 code.",
                                        "diag_3": "Tertiary diagnosis listed on each encounter, can be represented as an ICD10 code.",
                                        "diag_4": "Fourth diagnosis listed on each encounter, can be represented as an ICD10 code.",
                                        "diag_5": "Fifth diagnosis listed on each encounter, can be represented as an ICD10 code.",
                                        "proc_code": "Procedure performed within each encounter.",
                                        "proc_units": "Total number of procedure units within each encounter.",
                                        "proc_modifier": "Procedure code modifier for procedure performed within encounter.",
                                        "rev_code": "Revenue Center code designated for each encounter.",
                                        "rev_units": "Total number of days in which the revenue code was designated within the healthcare interaction.",
                                        "charge": "Line-item charge submitted to insurers. The charge is the pre-adjudicated value submitted to insurers.",
                                        "smart_allowed": "Line-item estimate of allowed amount, utilizing proprietary ML on CMS, Price Transparency and Actual Adjudicated amounts to provide estimates within an average of 5% variance from actuals."}
                            },
            "NPPES_TABLE":{"description":"This table is a cleaned, normalized and monthly-updated verions of the CMS NPPES database.", 
                           "queries":["nppesDetail"], 
                           "features":{"npi":"National Provider Identifier (NPI) of the individual/organizational provider registered with NPPES.",
                                       "name":"Full name or organizational name of the individual/organizational provider registered with NPPES.",
                                       "entity_type":"Indicator of whether the corresponding provider is an individual or organization.",
                                       "gender":"Gender of the individual provider.",
                                       "credential":"Medical credential of the individual provider.",
                                       "primary_taxonomy_code":"Primary taxonomy code of the individual/organizational provider.",
                                       "secondary_taxonomy_code":"Secondary taxonomy code of the individual/organizational provider.",
                                       "address":"Primary address of the individual/organizational provider, as reported to NPPES.",
                                       "metro":"City, State of the primary address of the individual/organizational provider, as reported to NPPES.",
                                       "state":"State of the primary address of the individual/organizational provider, as reported to NPPES.",
                                       "zip":"Zip code of the primary address of the individual/organizational provider, as reported to NPPES.",
                                       "phone":"Primary phone number of the individual/organizational provider, as reported to NPPES.",
                                       "fax":"Primary fax number of the individual/organizational provider, as reported to NPPES.",
                                       "sole_proprietor":"Binary indicator (yes, no) as to whether the provider is registered as a sole proprietor with NPPES.",
                                       "org_subpart":"Binary indicator (yes, no) as to whether the provider is registered as an organizational subpart with NPPES.",
                                       "forg_lbnax":"Legal business name of the parent organization associated with an NPI.",
                                       "update_date":"The last date in which the provider contact details have been updated"}
                            }, 
            "DIAG_TABLE":{"description":"This table provides helpful context and categories around ICD-10 diagnosis codes.", 
                           "queries":["diagDetail"], 
                           "features":{"diag_code":"ICD-10 diagnosis code",
                                       "short_diag_code":"First three characters of the ICD-10 diagnosis code, corresponding to the diagnosis subcategory",
                                       "diag_subcategory":"Name of the ICD-10 diagnosis sub-category",
                                       "diag_category":"Name of the ICD-10 diagnosis category",
                                       "diag_desc":"Description of the ICD-10 diagnosis code",
                                       "icd9_code":"ICD-9 mapping to the ICD-10 diagnosis code",
                                       "diag_severity":"CMS-designated severity score from 0 to 100"}
                            }, 
            "PROC_TABLE":{"description":"This table provides helpful context and categories around CPT/HCPCS procedure codes.", 
                           "queries":["procDetail"], 
                           "features":{"proc_code":"CPT/HCPCS procedural code",
                                       "proc_category":"Name of the CPT/HCPCS procedural category",
                                       "proc_subcategory":"Name of the CPT/HCPCS procedural sub-category",
                                       "proc_desc":"Description of the procedure code"}
                            },
            "SPECIALTY_TABLE":{"description":"This table provides helpful context and categories around taxonomy codes and specialty mappings.", 
                           "queries":["specDetail"], 
                           "features":{"taxonomy_code":"Taxonomy code classification system that identifies the specific area of practice of a healthcare provider",
                                       "subspecialty":"Sub-specialty are specific areas of focus within a healthcare specialty",
                                       "specialty":"Specialty is a branch of medicine that focuses on a specific area of the body, disease, or population",
                                       "specialty_category":"Broader categories of specialty areas of focus"}
                            },
            "MIPS_TABLE":{"description":"This table takes the monthly updates to the CMS MIPS database, cleans, normalizes and orients their output into a concise feature set built for consumption.", 
                           "queries":["mipsDetail"], 
                           "features":{"npi":"National Provider Identifier (NPI) of the individual/organizational provider registered with NPPES.",
                                       "org_pac_id":"Unique group ID assigned by PECOS to the group this individual participated in",
                                       "name":"Individual clinician's full name",
                                       "source":"Method by which the clinician achieved scores (“individual”, “group”, “virtual group”,  or “apm”); if the source is “apm”, this indicates that the MIPS APM scoring standard was applied for this set of scores",
                                       "quality_score":"Quality performance category score",
                                       "interop_score":"Promoting Interoperability (PI) performance category score",
                                       "improvement_score":"Improvement Activities (IA) performance category score",
                                       "mips_score_without_cpb":"MIPS Final Score without the Complex Patient Bonus (CPB). The CPB is part of the MIPS final score based on the overall medical complexity and social risk for one’s patients treated",
                                       "mips_score":"Final Merit-based Incentive Payment System (MIPS) score",
                                       "update_date":"The last date in which the provider contact details have been updated"}
                            },
            "GEO_TABLE":{"description":"This table allows provides a comprehensive view of every geographic area in the US, providing population and growth statistics from zip and county to geographic US regions.", 
                           "queries":["geoDetail"], 
                           "features":{"zip":"Postal Zone Improvement Plan (ZIP) code corresponding to a geographic area",
                                       "zip3":"First three digits of the zip code",
                                       "county_code":"US Government's Federal Information Processing Standards (FIPS) code for state-county combination",
                                       "county":"Name and State of the corresponding county",
                                       "county_pop":"US Census estimated population of the County population",
                                       "county_growth_rate":"US Census estimated County population growth trend",
                                       "metro":"US Census Bureau metropolitan and micropolitan statistical area",
                                       "metro_location":"Centrality of the county as it relates to the corresponding US Metro",
                                       "state_code":"Two digit code representing a US State",
                                       "state":"Full name of US State",
                                       "state_pop":"US Census estimated population of the State population",
                                       "state_growth_rate":"US Census estimated State population growth trend",
                                       "division":"Nine minor State groupings as defined by the US Census Bureau",
                                       "division_pop":"US Census estimated population of the Division population",
                                       "division_growth_rate":"US Census estimated Division population growth trend",
                                       "region":"Four major State groupings as defined by the US Census Bureau",
                                       "region_pop":"US Census estimated population of the Region population",
                                       "region_growth_rate":"US Census estimated Region population growth trend"}
                            },
            "EXEC_OUTREACH_TABLE":{"description":"This table provides methods of outreach for hospital and health system executives, including email, LinkedIn and role context.", 
                           "queries":["execOutreach"], 
                           "features":{"exec_name":"Full name of the executive record",
                                       "hospital":"Primary hospital/system associated with the executive record",
                                       "title":"Role title associated with the hospital/system executive record",
                                       "department":"Departmental affiliation with the role of the executive record",
                                       "exec_level":"Administrative level of the executive record",
                                       "email1":"Primary email address of the hospital/system executive record",
                                       "email2":"Secondary email address of the hospital/system executive record",
                                       "address":"Primary address of the hospital/system executive record",
                                       "city":"City for the primary address of the hospital/system executive record",
                                       "state":"US State code for the primary address of the hospital/system executive record",
                                       "zip":"Zip code for the primary address of the hospital/system executive record",
                                       "phone":"Primary phone number of the hospital/system executive record",
                                       "linkedin":"LinkedIn URL for the hospital/system executive record",
                                       "npi":"National Provider Identifier (NPI) if executive is an NPPES registered healthcare provider",
                                       "org_url":"Organizational website most closely associated with the executive record",
                                       "system_exec":"Biary indicator of whether the executive is at the hospital or health system administrative level",
                                       "update_date":"The last date in which the hospital executive contact details have been updated"}
                            }, 
            "PROV_OUTREACH_TABLE":{"description":"This table provides methods of outreach for individual healthcare providers, including emails, and primary locations.", 
                           "queries":["provOutreach"], 
                           "features":{"hcp_npi":"National Provider Identifier (NPI) of the individual provider registered with NPPES.",
                                       "hcp_name":"Full name of the individual provider record",
                                       "taxonomy":"Primary taxonomy for the provider record",
                                       "subspecialty":"Subspecialty name for the primary taxonomy of the provider record",
                                       "specialty":"Specialty name for the primary taxonomy of the provider record",
                                       "specialty_category":"Specialty category name for the primary taxonomy of the provider record",
                                       "email1":"Primary email address of the individual provider",
                                       "email2":"Secondary email address of the individual provider",
                                       "address":"Primary address on record of the individual provider",
                                       "city":"Primary city of address for the individual provider record",
                                       "state":"Two digit code representing a US State",
                                       "zip":"Postal Zone Improvement Plan (ZIP) code corresponding to a geographic area",
                                       "phone":"Primary phone number of the individual provider",
                                       "fax":"Primary fax number of the individual provider",
                                       "update_date":"The last date in which the provider contact details have been updated"}
                            },
            "AFFILIATION_TABLE":{"description":"This table offers comprehensive detail of individual provider affiliations through time, with particular attention to the totals of their actual clinical activity, where they practice, how those locations are affiliated with systems.", 
                           "queries":["affiliationActivity","updateAffiliation"], 
                           "features":{"hcp_npi":"National Provider Identifier (NPI) of the individual provider registered with NPPES.",
                                       "hcp_name":"Full name of the individual provider",
                                       "hcp_taxonomy":"Primary taxonomy code of the individual provider",
                                       "hcp_state":"US State code of the NPPES registered provider address",
                                       "hcp_short_zip":"Short zip code of the NPPES registered provider address",
                                       "system_affiliation":"Anticipated health system affiliation of the individual provider derived from their practice patterns and reported affiliation",
                                       "affiliation_confounded":"Binary indicator of whether provider practice patterns indicate multiple affiliations or absence of a reliable behavioral indicator of affiliation",
                                       "primary_hco_npi":"National Provider Identifier (NPI) of the facility/department where the individual provider performed the **most** encounters within one year of the indicated lookback year/quarter",
                                       "primary_hco_name":"Name of the facility/department where the individual provider performed the most encounters within one year of the indicated lookback year/quarter",
                                       "primary_hco_system":"Health system affiliation of the facility/department where the individual provider performed the most encounters within one year of the indicated lookback year/quarter",
                                       "primary_hco_encounters":"Number of encounters seen by individual provider at the primary_hco within the lookback window",
                                       "primary_hco_proportion":"Proportion of encounters seen by individual provider at the primary_hco compared to all other hco's, within the lookback window",
                                       "secondary_hco_npi":"National Provider Identifier (NPI) of the facility/department where the individual provider performed the second most encounters within one year of the indicated lookback year/quarter",
                                       "secondary_hco_name":"Name of the facility/department where the individual provider performed the second mostencounters within one year of the indicated lookback year/quarter",
                                       "secondary_hco_system":"Health system affiliation of the facility/department where the individual provider performed the second most encounters within one year of the indicated lookback year/quarter",
                                       "secondary_hco_encounters":"Number of encounters seen by individual provider at the secondary_hco within the lookback window",
                                       "secondary_hco_proportion":"Proportion of encounters seen by individual provider at the secondary_hco compared to all other hco's, within the lookback window",
                                       "tertiary_hco_npi":"National Provider Identifier (NPI) of the facility/department where the individual provider performed the third most encounters within one year of the indicated lookback year/quarter",
                                       "tertiary_hco_name":"Name of the facility/department where the individual provider performed the third most encounters within one year of the indicated lookback year/quarter",
                                       "tertiary_hco_system":"Health system affiliation of the facility/department where the individual provider performed the third most encounters within one year of the indicated lookback year/quarter",
                                       "tertiary_hco_encounters":"Number of encounters seen by individual provider at the tertiary_hco within the lookback window",
                                       "tertiary_hco_proportion":"Proportion of encounters seen by individual provider at the tertiary_hco compared to all other hco's, within the lookback window",
                                       "lookback_period":"End year-quarter value of the 1-year lookback period. Per the example, affiliation is calculated using encounters from the start of Q3-2022 to the start of Q3-2023",
                                       "current":"Binary indicator as to whether the period is the most recent recorded within CareQuery"}
                            },
            "ALLOWED_AVGS_TABLE":{"description":"This table gives users the ability to visualize the statistical shape of allowed amounts for procedure-specific claims across the US.",   
                           "queries":["procAllowedAvgs"], 
                           "features":{"proc_code": "CPT/HCPCS procedural code - used as a grouper for all corresponding statistics.",
                                       "proc_subcategory":"Sub-categorical description of the procedure code - used as a grouper for all corresponding statistics.",
                                       "state":"US State where the procedure was performed  - used as a grouper for all corresponding statistics.",
                                       "payor_channel":"Payor channel-of-business, line-of-business, or plan type  - used as a grouper for all corresponding statistics.",
                                       "visit_type": "Description of the setting of care  - used as a grouper for all corresponding statistics.",
                                       "total_encounters":"Total number of encounters corresponding to the groupers within the period and utilized in the allowed amount statistics.",
                                       "trimmed_mean_allowed_amount":"The mean allowed amount with outliers trimmed from consideration.",
                                       "mean_allowed_amount":"The mean allowed amount with all corresponding encounters considered.",
                                       "min_allowed_amount":"The minimum encounter allowed amount seen within the groupers/period.",
                                       "percentile_10":"The 10th percentile of all allowed amount values seen within the corresponding groupers/period.",
                                       "percentile_20":"The 20th percentile of all allowed amount values seen within the corresponding groupers/period.",
                                       "percentile_50":"The 50th percentile of all allowed amount values seen within the corresponding groupers/period.",
                                       "percentile_80":"The 80th percentile of all allowed amount values seen within the corresponding groupers/period.",
                                       "percentile_90":"The 90th percentile of all allowed amount values seen within the corresponding groupers/period.",
                                       "max_allowed_amount":"The maximum encounter allowed amount seen within the groupers/period.",
                                       "lookback_period":"All values are computed within an year-long lookback period on a quarterly basis. This value represents the most recent quarter-year in which a lookback has occurred.",
                                       "current":"Binary indicator as to whether the listed period is the most current.",}
                            },
            "SCRIPT_TABLE":{"description":"This table provides prescription claim with comprehensive drug and provider details, across entire longitudinal journeys.",   
                           "queries":["scriptEncounter", "scriptJourney"], 
                           "features":{"journey_id":"Randomly generated identifier unique to each longitudinal patient journey",
                                       "script_id":"Randomly generated identifier unique to each prescription claim",
                                       "group_id":"Randomly generated identifier unique to each payor plan corresponding to the prescription written",
                                       "script_date":"Date the prescription was written ",
                                       "service_date":"Date that the prescription was filled, services rendered, or the payer began coverage following Part A expiration in a long-term care setting",
                                       "patient_state":"US State residence of the patient on each encounter",
                                       "patient_short_zip":"First three digits of the patient zip code",
                                       "patient_age":"Age of the patient at the moment of each encounter",
                                       "patient_gender":"Gender/Sex of the patient",
                                       "hcp_npi":"National Provider Identifier (NPI) of the individual that wrote the prescription for the patient",
                                       "hcp_taxonomy":"Primary taxonomy code of the individual provider who wrote the prescription",
                                       "hcp_specialty":"Primary specialty of the individual provider provider wrote the prescription",
                                       "hco_npi":"National Provider Identifier (NPI) of the organization that fulfilled the prescription",
                                       "ndc":"FDA National Drug Code version 11 standardized code for a dispensed drug",
                                       "proprietary_name":"Unique brand name of a drug product as assigned by the drug manufacturer",
                                       "refills":"Maximum number of refills authorized",
                                       "quantity_prescribed":"Quantity of drug units prescribed",
                                       "quantity_dispensed":"Quantity of drug units dispensed",
                                       "fill_number":"Refill number for the current SCRIPT_ID",
                                       "days_supply":"Estimated number of days the dispensed drugs will last",
                                       "unit_of_measure":"Units of measure for the quantity_dispensed",
                                       "submitted_cost":"Pharmacy submitted cost of the claim",
                                       "patient_pay":"Patient cost responsibility as determined by payor",
                                       "copay_insurance":"Patient cost responsibility as copay or coinsurance",
                                       "plan_pay":"Payor cost responsibility",}
                            },
            "SCRIPT_DETAIL_TABLE":{"description":"This table provides ingredient, company, and categorical detail for all FDA National Drug Codes (NDCs).",   
                           "queries":["scriptDetail"], 
                           "features":{"ndc":"FDA standardized National Drug Code 11 for a dispensed drug",
                                       "labeler":"A company or organization that is a manufacturer, repackager, relabeler, or distributor of drugs",
                                       "product_type":"A classification of a drug product based on its intended use and how it is regulated by the FDA",
                                       "substance":"A substance is the active ingredient in a drug product",
                                       "substance_unit":"The unit of measurement for the active ingredient in a drug product",
                                       "dosage_form":"The type of pharmaceutical formulation, such as a tablet, capsule, liquid, or injection, that a drug product is available in",
                                       "route":"Method by which a drug product is administered to the body",
                                       "marketing_category":"A classification of a drug product based on its regulatory status and how it is marketed",
                                       "application_number":"A unique identifier for an NDC labeler's request to obtain an NDC labeler code",
                                       "pharm_class":"A list of classifications of a drug product based on its therapeutic action, or how it works in the body",}
                            },
            "DRG_TABLE":{"description":"This table provides helpful context and categories around Diagnosis Related Groups (DRGs) for inpatient medical billing.",   
                           "queries":["drgDetail"], 
                           "features":{"drg_code":"Diagnosis related group (DRG) code. This code is based on your primary and secondary diagnoses, other conditions (comorbidities), age, sex, and necessary medical procedures",
                                       "drg_name":"Descriptive name of the DRG code",
                                       "drg_category":"Medical Diagnostic Category MDC of the DRG",
                                       "drg_type":"The type of diagnosis related group (DRG)",
                                       "drg_post_acute":"Binary indicator of whether a DRG is post acute or not",
                                       "drg_special":"Binary indicator of whether a DRG is a special condition - particularly complex or expensive conditions that  may require specialized care, longer hospital stays, and/or more expensive medications and procedures",
                                       "drg_weight":"Represents the relative resources required  to care for patients in that particular DRG. This CMS computed figure highlights a DRG's relative to the average resources used to treat cases in all other DRGs",
                                       "los_gmean":"The geometric mean length of stay corresponding to a DRG",
                                       "los_mean":"The arithmetic mean length of stay corresponding to a DRG",}
                            },
        }

queries = {"careEncounter":{"description" : "Specify encounter-based query parameters to construct a query object. From which the .execute() method on this query will return line-item level patient journey detail for healthcare encounters that meet the query criterion. We define an encoutner as a single, billiable healthcare instance of care.",
                            "parameters" : {"min_date" : "Lower threshold date within the date range of interest, observational start date",
                                            "max_date" : "Upper threshold date within the date range of interest, observational end date",
                                            "gender" : "Patient gender within the population of interest",
                                            "age_min" : "Minimum patient age identified within the population of interest",
                                            "age_max" : "Maximum patient age identified within the population of interest",
                                            "visit_type" : "Description of the setting of care",
                                            "payor_channel" : "Health Insurance Payor channel-of-business or line-of-business",
                                            "revenue_center_code" : "4-digit code used to identify a specific accommodation, ancillary service, or billing calculation on a hospital bill",
                                            "place_of_service" : "Setting in which a healthcare service was provided by healthcare providers, insurers, and government agencies to track and bill for healthcare services",
                                            "drg_code":"Diagnosis related group (DRG) code. This code is based on your primary and secondary diagnoses, other conditions (co-morbidities), age, sex, and necessary medical procedures.",
                                            "drg_category":"Medical Diagnostic Category (MDC) of the Diagnosis Related Group (DRG)",
                                            "drg_greedy":"Boolean indicator for 'AND' or 'OR' when multiple DRG parameters specified - helps users offer various levels of specificity without confusing and/or logic",
                                            "diag_code" : "ICD-10 diagnosis code assigned to the patient",
                                            "diag_category" : "ICD-10 diagnosis category assigned to the patient",
                                            "diag_subcategory" : "ICD-10 diagnosis subcategory assigned to the patient",
                                            "short_diag_code" : "First three digits of ICD-10 code assigned to the patient",
                                            "diag_greedy" : "Boolean indicator for 'AND' or 'OR' when multiple diagnosis parameters specified - helps users offer various levels of specificity without confusing and/or logic",
                                            "proc_code" : "CPT or HCPCS procedural code performed",
                                            "proc_category" : "CPT or HCPCS procedural category performed",
                                            "proc_subcategory" : "CPT or HCPCS procedural subcategory performed",
                                            "proc_greedy" : "Boolean indicator for 'AND' or 'OR' when multiple procedure parameters specified - helps users offer various levels of specificity without confusing and/or logic",
                                            "taxonomy_code" : "Taxonomy of the organization or individual providers, 10-digit code that used to identify the type, classification, and area of specialization of an individual or organizational healthcare provider",
                                            "specialty" : "Specialty of the organization or individual provider(s), specific focus of a provider by patient population, disease, skill or philosophy",
                                            "specialty_category" : "Specialty category of the organization or individual provider(s), broad focus of a provider by patient population, disease, skill or philosophy",
                                            "subspecialty" : "Specialty subcategory of the organization or individual provider(s), hyper-specific focus of a provider by patient population, disease, skill or philosophy",
                                            "taxonomy_greedy" : "Boolean indicator for 'AND' or 'OR' when multiple taxonomy or specialty parameters specified - helps users offer various levels of specificity without confusing and/or logic",
                                            "ref_npi":"Referring provider(s) indicated by National Provider Identifiers (NPIs)",
                                            "npi":"National Provider Identifier (NPI) of individual and organizational provider(s) of interest",
                                            "state":"US State residence of the query patient population",
                                            "division":"Nine distinct US State groupings as defined by the US Census Bureau",
                                            "region":"Four major US State groupings as defined by the US Census Bureau - Northeast, Midwest, South and West",
                                            "metro":"Metropolitan and micropolitan statistical areas according to the US Census Bureau",
                                            "short_zip":"First three digits the zip code of the query patient population specified",
                                            "limit":"Number of rows returned in the data request.",
                                            "show":"Binary indicator as to whether you'd like your query to be printed for review upon it's creation. Regardless, this summary may be accessed via the .summarize(show=True) method on the query object itself."}
                        },
            "careVisit":{"description" : "Visit-based query that returns line-item level patient journey detail for healthcare visits that meet the query criterion. We define a visit as group of encounters within the same event of of care.",
                            "parameters" : {"min_date" : "Lower threshold date within the date range of interest, observational start date",
                                            "max_date" : "Upper threshold date within the date range of interest, observational end date",
                                            "gender" : "Patient gender within the population of interest",
                                            "age_min" : "Minimum patient age identified within the population of interest",
                                            "age_max" : "Maximum patient age identified within the population of interest",
                                            "visit_type" : "Description of the setting of care",
                                            "payor_channel" : "Health Insurance Payor channel-of-business or line-of-business",
                                            "revenue_center_code" : "4-digit code used to identify a specific accommodation, ancillary service, or billing calculation on a hospital bill",
                                            "place_of_service" : "Setting in which a healthcare service was provided by healthcare providers, insurers, and government agencies to track and bill for healthcare services",
                                            "diag_code" : "ICD-10 diagnosis code assigned to the patient",
                                            "diag_category" : "ICD-10 diagnosis category assigned to the patient",
                                            "diag_subcategory" : "ICD-10 diagnosis subcategory assigned to the patient",
                                            "short_diag_code" : "First three digits of ICD-10 code assigned to the patient",
                                            "diag_greedy" : "Boolean indicator for 'AND' or 'OR' when multiple diagnosis parameters specified - helps users offer various levels of specificity without confusing and/or logic",
                                            "proc_code" : "CPT or HCPCS procedural code performed",
                                            "proc_category" : "CPT or HCPCS procedural category performed",
                                            "proc_subcategory" : "CPT or HCPCS procedural subcategory performed",
                                            "proc_greedy" : "Boolean indicator for 'AND' or 'OR' when multiple procedure parameters specified - helps users offer various levels of specificity without confusing and/or logic",
                                            "taxonomy_code" : "Taxonomy of the organization or individual providers, 10-digit code that used to identify the type, classification, and area of specialization of an individual or organizational healthcare provider",
                                            "specialty" : "Specialty of the organization or individual provider(s), specific focus of a provider by patient population, disease, skill or philosophy",
                                            "specialty_category" : "Specialty category of the organization or individual provider(s), broad focus of a provider by patient population, disease, skill or philosophy",
                                            "subspecialty" : "Specialty subcategory of the organization or individual provider(s), hyper-specific focus of a provider by patient population, disease, skill or philosophy",
                                            "taxonomy_greedy" : "Boolean indicator for 'AND' or 'OR' when multiple taxonomy or specialty parameters specified - helps users offer various levels of specificity without confusing and/or logic",
                                            "ref_npi":"Referring provider(s) indicated by National Provider Identifiers (NPIs)",
                                            "npi":"National Provider Identifier (NPI) of individual and organizational provider(s) of interest",
                                            "state":"US State residence of the query patient population",
                                            "division":"Nine distinct US State groupings as defined by the US Census Bureau",
                                            "region":"Four major US State groupings as defined by the US Census Bureau - Northeast, Midwest, South and West",
                                            "metro":"Metropolitan and micropolitan statistical areas according to the US Census Bureau",
                                            "short_zip":"First three digits the zip code of the query patient population specified",
                                            "limit":"Number of rows returned in the data request.",
                                            "show":"Binary indicator as to whether you'd like your query to be printed for review upon it's creation. Regardless, this summary may be accessed via the .summarize(show=True) method on the query object itself."}
                            },
            "careEpisode":{"description" : "Episode-based query that returns line-item level patient journey detail for healthcare episodes that meet the query criterion. We define an episode as sequential group of encounters that occur within the 30 days from one another, in continuation, until greater than 30 days has elapsed.",
                            "parameters" : {"min_date" : "Lower threshold date within the date range of interest, observational start date",
                                            "max_date" : "Upper threshold date within the date range of interest, observational end date",
                                            "gender" : "Patient gender within the population of interest",
                                            "age_min" : "Minimum patient age identified within the population of interest",
                                            "age_max" : "Maximum patient age identified within the population of interest",
                                            "visit_type" : "Description of the setting of care",
                                            "payor_channel" : "Health Insurance Payor channel-of-business or line-of-business",
                                            "revenue_center_code" : "4-digit code used to identify a specific accommodation, ancillary service, or billing calculation on a hospital bill",
                                            "place_of_service" : "Setting in which a healthcare service was provided by healthcare providers, insurers, and government agencies to track and bill for healthcare services",
                                            "diag_code" : "ICD-10 diagnosis code assigned to the patient",
                                            "diag_category" : "ICD-10 diagnosis category assigned to the patient",
                                            "diag_subcategory" : "ICD-10 diagnosis subcategory assigned to the patient",
                                            "short_diag_code" : "First three digits of ICD-10 code assigned to the patient",
                                            "diag_greedy" : "Boolean indicator for 'AND' or 'OR' when multiple diagnosis parameters specified - helps users offer various levels of specificity without confusing and/or logic",
                                            "proc_code" : "CPT or HCPCS procedural code performed",
                                            "proc_category" : "CPT or HCPCS procedural category performed",
                                            "proc_subcategory" : "CPT or HCPCS procedural subcategory performed",
                                            "proc_greedy" : "Boolean indicator for 'AND' or 'OR' when multiple procedure parameters specified - helps users offer various levels of specificity without confusing and/or logic",
                                            "taxonomy_code" : "Taxonomy of the organization or individual providers, 10-digit code that used to identify the type, classification, and area of specialization of an individual or organizational healthcare provider",
                                            "specialty" : "Specialty of the organization or individual provider(s), specific focus of a provider by patient population, disease, skill or philosophy",
                                            "specialty_category" : "Specialty category of the organization or individual provider(s), broad focus of a provider by patient population, disease, skill or philosophy",
                                            "subspecialty" : "Specialty subcategory of the organization or individual provider(s), hyper-specific focus of a provider by patient population, disease, skill or philosophy",
                                            "taxonomy_greedy" : "Boolean indicator for 'AND' or 'OR' when multiple taxonomy or specialty parameters specified - helps users offer various levels of specificity without confusing and/or logic",
                                            "ref_npi":"Referring provider(s) indicated by National Provider Identifiers (NPIs)",
                                            "npi":"National Provider Identifier (NPI) of individual and organizational provider(s) of interest",
                                            "state":"US State residence of the query patient population",
                                            "division":"Nine distinct US State groupings as defined by the US Census Bureau",
                                            "region":"Four major US State groupings as defined by the US Census Bureau - Northeast, Midwest, South and West",
                                            "metro":"Metropolitan and micropolitan statistical areas according to the US Census Bureau",
                                            "short_zip":"First three digits the zip code of the query patient population specified",
                                            "limit":"Number of rows returned in the data request.",
                                            "show":"Binary indicator as to whether you'd like your query to be printed for review upon it's creation. Regardless, this summary may be accessed via the .summarize(show=True) method on the query object itself."}
                        },
        "careJourney":{"description" : "Journey-based query that returns line-item level patient journey detail for healthcare patient journeys that meet the query criterion. We define a journey as a full longitudinal patient record, all encounters dating back to January 1st, 2019.",
                            "parameters" : {"min_date" : "Lower threshold date within the date range of interest, observational start date",
                                            "max_date" : "Upper threshold date within the date range of interest, observational end date",
                                            "gender" : "Patient gender within the population of interest",
                                            "age_min" : "Minimum patient age identified within the population of interest",
                                            "age_max" : "Maximum patient age identified within the population of interest",
                                            "visit_type" : "Description of the setting of care",
                                            "payor_channel" : "Health Insurance Payor channel-of-business or line-of-business",
                                            "revenue_center_code" : "4-digit code used to identify a specific accommodation, ancillary service, or billing calculation on a hospital bill",
                                            "place_of_service" : "Setting in which a healthcare service was provided by healthcare providers, insurers, and government agencies to track and bill for healthcare services",
                                            "diag_code" : "ICD-10 diagnosis code assigned to the patient",
                                            "diag_category" : "ICD-10 diagnosis category assigned to the patient",
                                            "diag_subcategory" : "ICD-10 diagnosis subcategory assigned to the patient",
                                            "short_diag_code" : "First three digits of ICD-10 code assigned to the patient",
                                            "diag_greedy" : "Boolean indicator for 'AND' or 'OR' when multiple diagnosis parameters specified - helps users offer various levels of specificity without confusing and/or logic",
                                            "proc_code" : "CPT or HCPCS procedural code performed",
                                            "proc_category" : "CPT or HCPCS procedural category performed",
                                            "proc_subcategory" : "CPT or HCPCS procedural subcategory performed",
                                            "proc_greedy" : "Boolean indicator for 'AND' or 'OR' when multiple procedure parameters specified - helps users offer various levels of specificity without confusing and/or logic",
                                            "taxonomy_code" : "Taxonomy of the organization or individual providers, 10-digit code that used to identify the type, classification, and area of specialization of an individual or organizational healthcare provider",
                                            "specialty" : "Specialty of the organization or individual provider(s), specific focus of a provider by patient population, disease, skill or philosophy",
                                            "specialty_category" : "Specialty category of the organization or individual provider(s), broad focus of a provider by patient population, disease, skill or philosophy",
                                            "subspecialty" : "Specialty subcategory of the organization or individual provider(s), hyper-specific focus of a provider by patient population, disease, skill or philosophy",
                                            "taxonomy_greedy" : "Boolean indicator for 'AND' or 'OR' when multiple taxonomy or specialty parameters specified - helps users offer various levels of specificity without confusing and/or logic",
                                            "ref_npi":"Referring provider(s) indicated by National Provider Identifiers (NPIs)",
                                            "npi":"National Provider Identifier (NPI) of individual and organizational provider(s) of interest",
                                            "state":"US State residence of the query patient population",
                                            "division":"Nine distinct US State groupings as defined by the US Census Bureau",
                                            "region":"Four major US State groupings as defined by the US Census Bureau - Northeast, Midwest, South and West",
                                            "metro":"Metropolitan and micropolitan statistical areas according to the US Census Bureau",
                                            "short_zip":"First three digits the zip code of the query patient population specified",
                                            "limit":"Number of rows returned in the data request.",
                                            "show":"Binary indicator as to whether you'd like your query to be printed for review upon it's creation. Regardless, this summary may be accessed via the .summarize(show=True) method on the query object itself."}
                        },
        "nppesDetail":{"description" : "Query to navigate the NPPES_TABLE, which sources monthly updates to the national provider database within the National Plan & Provider Enumeration System (NPPES) by CMS.",
                        "parameters" : {"npi":"National Provider Identifier (NPI) of individual and organizational provider(s) of interest",
                                        "taxonomy_code":"Taxonomy of the organization or individual providers, 10-digit code that used to identify the type, classification, and area of specialization of an individual or organizational healthcare provider",
                                        "specialty":"Specialty of the organization or individual provider(s), specific focus of a provider by patient population, disease, skill or philosophy",
                                        "specialty_category":"Specialty category of the organization or individual provider(s), broad focus of a provider by patient population, disease, skill or philosophy",
                                        "subspecialty":"Specialty subcategory of the organization or individual provider(s), hyper-specific focus of a provider by patient population, disease, skill or philosophy",
                                        "state":"US State of the provider's NPPES registered address",
                                        "division":"US State Division of the provider's NPPES registered address. There are nine distinct US State groupings as defined by the US Census Bureau",
                                        "region":"US State Region of the provider's NPPES registered address. There are four major US State groupings as defined by the US Census Bureau",
                                        "zip":"Primary zip code of the provider",
                                        "limit":"Number of rows returned in the data request",
                                        "show":"Binary indicator as to whether you'd like your query to be printed for review upon it's creation. Regardless, this summary may be accessed via the .summarize(show=True) method on the query object itself."}
                        },
        "diagDetail":{"description" : "Query to specify the diagnosis parameters for additional context, categories and code ties from the DIAG_TABLE",
                        "parameters" : {"diag_code":"ICD-10 diagnosis code",
                                        "diag_category":"Name of the ICD-10 diagnosis category",
                                        "diag_subcategory":"Name of the ICD-10 diagnosis sub-category",
                                        "short_diag_code":"First three characters of the ICD-10 diagnosis code, corresponding to the diagnosis subcategory",
                                        "limit":"Number of rows returned in the data request",
                                        "show":"Binary indicator as to whether you'd like your query to be printed for review upon it's creation. Regardless, this summary may be accessed via the .summarize(show=True) method on the query object itself"}
                        },
        "procDetail":{"description" : "Query to specify the procedure parameters for additional context, categories and code ties from the PROC_TABLE",
                        "parameters" : {"proc_code":"CPT/HCPCS procedural code",
                                        "proc_category":"Name of the CPT/HCPCS procedural category",
                                        "proc_subcategory":"Name of the CPT/HCPCS procedural sub-category",
                                        "limit":"Number of rows returned in the data request",
                                        "show":"Binary indicator as to whether you'd like your query to be printed for review upon it's creation. Regardless, this summary may be accessed via the .summarize(show=True) method on the query object itself"}
                        },
        "specialtyDetail":{"description" : "Query to specify the taxonmy/specialty parameters for additional context, categories and code ties from the SPEC_TABLE",
                        "parameters" : {"taxonomy_code":"Taxonomy of organization or individual provider, 10-digit code that used to identify the type, classification, and area of specialization of an individual or organizational healthcare provider",
                                        "subspecialty":"Specialty subcategory of the organization or individual provider(s), hyper-specific focus of a provider by patient population, disease, skill or philosophy",
                                        "specialty":"Specialty of organization or individual provider(s), specific focus of a provider by patient population, disease, skill or philosophy",
                                        "specialty_category":"Specialty category of the organization or individual provider(s), broad focus of a provider by patient population, disease, skill or philosophy",
                                        "limit":"Number of rows returned in the data request",
                                        "show":"Binary indicator as to whether you'd like your query to be printed for review upon it's creation. Regardless, this summary may be accessed via the .summarize(show=True) method on the query object itself"}
                        },
        "mipsDetail":{"description" : "Query to request Merit-based Incentive Payment System (MIPS) performance information for specified providers",
                        "parameters" : {"npi":"National Provider Identifier (NPI) assigned to the individual provider",
                                        "limit":"Number of rows returned in the data request",
                                        "show":"Binary indicator as to whether you'd like your query to be printed for review upon it's creation. Regardless, this summary may be accessed via the .summarize(show=True) method on the query object itself"}
                        },
        "geoDetail":{"description" : "Query to specify geographic detail around zip, county, metro, state or region - including US Census population statistics at each level",
                        "parameters" : {"short_zip":"First three digits any valid ZIP code",
                                        "metro":"Metropolitan and micropolitan statistical areas according to the US Census Bureau",
                                        "state":"Two digit US State code(s)",
                                        "division":"Nine distinct US State groupings as defined by the US Census Bureau",
                                        "region":"Four major US State groupings as defined by the US Census Bureau",
                                        "limit":"Number of rows returned in the data request",
                                        "show":"Binary indicator as to whether you'd like your query to be printed for review upon it's creation. Regardless, this summary may be accessed via the .summarize(show=True) method on the query object itself"}
                        },
        "execOutreach":{"description" : "Query to specify executive email, phone and address data - including context around their role and seniority.",
                        "parameters" : {"hospital":"Name of the primary hospital affiliated with the executive",
                                        "exec_level":"Adiministrative level associated with the executive's role",
                                        "department":"Functional department associated with the executive's role",
                                        "npi":"Where executive is a provider, their National Provider Identifier (NPI)",
                                        "state":"US State code of the executive record",
                                        "division":"US State Division code of the executive record",
                                        "region":"US State Region code of the executive record",
                                        "metro":"Metropolitan and micropolitan statistical areas according to the US Census Bureau",
                                        "short_zip":"First three digits of zip code for the executive record",
                                        "limit":"Number of rows returned in the data request",
                                        "show":"Binary indicator as to whether you'd like your query to be printed for review upon it's creation. Regardless, this summary may be accessed via the .summarize(show=True) method on the query object itself"}
                        },
        "providerOutreach":{"description" : "Query to specify individual healthcare providers and their accompanying contact information, including email, phone and fax.",
                        "parameters" : {"npi":"National Provider Identifier (NPI) of individual and organizational provider record",
                                        "taxonomy_code":"Primary taxonomy code of the individual provider record",
                                        "subspecialty":"Subspecialty of the individual provider record",
                                        "specialty":"Specialty of the individual provider record",
                                        "specialty_category":"Specialty category of the individual provider record",
                                        "state":"US State code of the individual provider record",
                                        "division":"US State Division code of the individual provider record",
                                        "region":"US State Region code of the individual provider record",
                                        "metro":"Metropolitan and micropolitan statistical areas according to the US Census Bureau",
                                        "short_zip":"First three digits of zip code for the individual provider record",
                                        "limit":"Number of rows returned in the data request",
                                        "show":"Binary indicator as to whether you'd like your query to be printed for review upon it's creation. Regardless, this summary may be accessed via the .summarize(show=True) method on the query object itself"}
                        },
         "affiliationActivity":{"description" : "Query to specify health system provider affiliation query parameters to return nuanced affiliation data, including the actual claims activity-by-affiliation for each provider",
                        "parameters" : {"npi":"National Provider Identifier (NPI) of individual and organizational provider record",
                                        "system":"Name of the system an individual provider may be affiliated with in any way - either indicated by overall affiliation, or via the system affiliation for any of the top 3 healthcare organization (hco) in which they practice at",
                                        "taxonomy_code":"Primary taxonomy code of the individual provider record",
                                        "subspecialty":"Subspecialty of the individual provider record",
                                        "specialty":"Specialty of the individual provider record",
                                        "specialty_category":"Specialty category of the individual provider record",
                                        "state":"US State code of the individual provider record",
                                        "division":"US State Division code of the individual provider record",
                                        "region":"US State Region code of the individual provider record",
                                        "lookback_period":"String indicating the quarter and year of the provider practice",
                                        "limit":"Number of rows returned in the data request",
                                        "show":"Binary indicator as to whether you'd like your query to be printed for review upon it's creation. Regardless, this summary may be accessed via the .summarize(show=True) method on the query object itself"}
                        },
        "affiliationActivity":{"description" : "Query to update health system and provider affiliations given provider/system parameters given",
                        "parameters" : {"system":"Name of the system an individual provider may be affiliated with in any way - either indicated by overall affiliation, or via the system affiliation for any of the top 3 healthcare organization (hco) in which they practice at",
                                        "add_npi":"National Provider Identifier (NPI) an individual and/or organizational provider(s) with which you would like to ADD to the affiliation list of the system specified",
                                        "remove_npi":"National Provider Identifier (NPI) an individual and/or organizational provider(s) with which you would like to REMOVE from the affiliation list of the system specified",
                                        "show":"Binary indicator as to whether you'd like your query to be printed for review upon it's creation. Regardless, this summary may be accessed via the .summarize(show=True) method on the query object itself"}
                        },
        "procAllowedAvgs":{"description" : "Query to specify procedure-encounter-based criterion and return statistical data around healthcare claim allowed amounts.",
                            "parameters" : {"visit_type" : "Description of the setting of care",
                                            "payor_channel" : "Health Insurance Payor channel-of-business or line-of-business",
                                            "proc_code" : "CPT or HCPCS procedural code performed",
                                            "proc_category" : "CPT or HCPCS procedural category performed",
                                            "proc_subcategory" : "CPT or HCPCS procedural subcategory performed",
                                            "state":"US State residence of the query patient population",
                                            "division":"Nine distinct US State groupings as defined by the US Census Bureau",
                                            "region":"Four major US State groupings as defined by the US Census Bureau - Northeast, Midwest, South and West",
                                            "limit":"Number of rows returned in the data request.",
                                            "show":"Binary indicator as to whether you'd like your query to be printed for review upon it's creation. Regardless, this summary may be accessed via the .summarize(show=True) method on the query object itself."}
                        },
        "scriptEncounter":{"description" : "Query patient prescription claims that meet the specified criterion with line item detail",
                            "parameters" : {"min_date":"Lower threshold date within the date range of interest, observational start date",
                                            "max_date":"Upper threshold date within the date range of interest, observational end date",
                                            "":"",
                                            "gender":"Patient gender within the population of interest",
                                            "age_min":"Minimum patient age identified within the population of interest",
                                            "age_max":"Maximum patient age identified within the population of interest",
                                            "ndc":"FDA National Drug Code version 11 standardized code for a dispensed drug",
                                            "pharm_class":"Classifications of a drug product based on its therapeutic action, or how it works in the body",
                                            "substance":"The active ingredient in a drug product",
                                            "marketing_category":"Classification of a drug product based on its regulatory status and how it is marketed",
                                            "taxonomy_code":"Taxonomy of the organization or individual provider, 10-digit code that used to identify the type, classification, and area of specialization of an individual or organizational healthcare provider",
                                            "specialty":"Specialty of the organization or individual provider(s), specific focus of a provider by patient population, disease, skill or philosophy",
                                            "specialty_category":"Specialty category of the organization or individual provider(s), broad focus of a provider by patient population, disease, skill or philosophy",
                                            "subspecialty":"Specialty subcategory of the organization or individual provider(s), hyper-specific focus of a provider by patient population, disease, skill or philosophy",
                                            "npi":"National Provider Identifier (NPI) of individual and/or organizational provider(s) that appear on the encounter",
                                            "state":"US State residence of the query patient population",
                                            "division":"Nine distinct US State groupings as defined by the US Census Bureau",
                                            "region":"Four major US State groupings as defined by the US Census Bureau",
                                            "metro":"Metropolitan and micropolitan statistical  \nareas according to the US Census Bureau",
                                            "short_zip":"First three digits the zip code of the  \nquery patient population specified",
                                            "limit":"Number of rows returned in the data request",
                                            "show":"Binary indicator as to whether you'd like  \nquery to be printed for review upon it's creation"}
                        },
        "scriptJourney":{"description" : "Query patient's full prescription claims history with line item detail, by specifying ANY and ALL prescription parameters where criterion are met across the entire journey",
                            "parameters" : {"min_date":"Lower threshold date within the date range of interest, observational start date",
                                            "max_date":"Upper threshold date within the date range of interest, observational end date",
                                            "gender":"Patient gender within the population of interest",
                                            "age_min":"Minimum patient age identified within the population of interest",
                                            "age_max":"Maximum patient age identified within the population of interest",
                                            "ndc":"FDA National Drug Code version 11 standardized code for a dispensed drug",
                                            "pharm_class":"Classifications of a drug product based on its therapeutic action, or how it works in the body",
                                            "substance":"The active ingredient in a drug product",
                                            "marketing_category":"Classification of a drug product based on its regulatory status and how it is marketed",
                                            "taxonomy_code":"Taxonomy of the organization or individual provider, 10-digit code that used to identify the type, classification, and area of specialization of an individual or organizational healthcare provider",
                                            "specialty":"Specialty of the organization or individual provider(s), specific focus of a provider by patient population, disease, skill or philosophy",
                                            "specialty_category":"Specialty category of the organization or individual provider(s), broad focus of a provider by patient population, disease, skill or philosophy",
                                            "subspecialty":"Specialty subcategory of the organization or individual provider(s), hyper-specific focus of a provider by patient population, disease, skill or philosophy",
                                            "npi":"National Provider Identifier (NPI) of individual and/or organizational provider(s) that appear on the encounter",
                                            "state":"US State residence of the query patient population",
                                            "division":"Nine distinct US State groupings as defined by the US Census Bureau",
                                            "region":"Four major US State groupings as defined by the US Census Bureau",
                                            "metro":"Metropolitan and micropolitan statistical  \nareas according to the US Census Bureau",
                                            "short_zip":"First three digits the zip code of the  \nquery patient population specified",
                                            "limit":"Number of rows returned in the data request",
                                            "show":"Binary indicator as to whether you'd like  \nquery to be printed for review upon it's creation"}
                        },
        "scriptDetail":{"description" : "Query prescription details such as active ingredient, company labeler, and categorical details for all FDA National Drug Codes (NDCs).",
                            "parameters" : {"ndc":"FDA National Drug Code version 11 standardized code for a dispensed drug",
                                            "pharm_class":"Classifications of a drug product based on its therapeutic action, or how it works in the body",
                                            "substance":"The active ingredient in a drug product",
                                            "marketing_category":"Classification of a drug product based on its regulatory status and how it is marketed",
                                            "limit":"Number of rows returned in the data request",
                                            "show":"Binary indicator as to whether you'd like  \nquery to be printed for review upon it's creation"}
                        },
        "drgDetail":{"description" : "Query to specify data request for all Diagnosis Related Group (DRG) parameters for additional context and categories.",
                            "parameters" : {"drg_code":"Diagnosis related group (DRG) code. This code is based on your primary and secondary diagnoses, other conditions (co-morbidities), age, sex, and necessary medical procedures.",
                                            "drg_category":"Medical Diagnostic Category (MDC) of the Diagnosis Related Group (DRG)",
                                            "limit":"Number of rows returned in the data request",
                                            "show":"Binary indicator as to whether you'd like  \nquery to be printed for review upon it's creation"}
                        },
        }

### Table Feature Data Class ###
class DataGlossary:

    def __init__(self):
        
        self.tables = tables
        self.queries = queries

        # read field value json - move to CQ_Class or CQUtils
        #FIXME: Uncomment and use this for local testing
        # with open('field-values.json', 'r') as f:
        #     self.parameter_options = json.load(f)

        #FIXME: Use this for the pypi distro
        with importlib.resources.open_text('care_query', 'field-values.json') as f:
            self.parameter_options = json.load(f)


    def list_tables(self):

        listed = list(self.tables.keys())

        return listed
    
    def list_queries(self):

        listed = list(self.queries.keys())

        return listed
    
    def list_parameters(self, query):

        listed = self.queries[query]["parameters"]

        return listed
    
    def list_options(self, parameter, contains = None):

        try:
            if type(self.parameter_options[parameter]) == dict:
                listed = list(self.parameter_options[parameter].keys())
                if contains != None:
                    listed = [x for x in listed if contains in str(x).lower()]
            elif type(self.parameter_options[parameter]) == list:
                listed = self.parameter_options[parameter]
                if contains != None:
                    listed = [x for x in listed if contains in str(x).lower()]
            else:
                listed = []
            if (len(listed) == 0) & (contains != None):
                print("No parameter options meet the contains critera.")
        except:
            raise ValueError(f"""ValueError: {parameter} parameter Not Found. Run self.list_parameters("name of query") to see all parameter options""")

        return listed
    
    def parameter_detail(self, parameter, as_pandas = None):
        
        try:
            detail = self.parameter_options[parameter]
            # if as_pandas != False:
            #     print(pd.DataFrame.from_dict(detail))
            #     #detail.columns = ["parameter","options"]
        except:
            raise ValueError(f"""ValueError: {parameter} parameter Not Found. Run self.list_parameters("name of query") to see all parameter options""")
        
        return detail

    def table_detail(self, table):

        try:
            detail = self.tables[table]
        except:
            raise ValueError(f"ValueError: {table} table Not Found. Run self.list_tables() to see all table options")

        return detail

    def feature_detail(self, table):

        try:
            detail = self.tables[table]["features"]
        except:
            raise ValueError(f"ValueError: {table} table Not Found. Run self.list_tables() to see all table options")

        return detail

    def query_detail(self, query):

        try:
            detail = self.queries[query]
        except:
            raise ValueError(f"ValueError: {query} method Not Found. Run self.list_methods() to see all method options")

        return detail